from typing import TYPE_CHECKING

from slixmpp import JID, CoroutineCallback, Iq, StanzaPath
from slixmpp.exceptions import XMPPError

from ...util.db import user_store

if TYPE_CHECKING:
    from .base import BaseGateway


class Search:
    def __init__(self, xmpp: "BaseGateway"):
        self.xmpp = xmpp

        xmpp["xep_0055"].api.register(self.search_get_form, "search_get_form")
        xmpp["xep_0055"].api.register(self._search_query, "search_query")

        xmpp.plugin["xep_0030"].add_feature("jabber:iq:gateway")
        xmpp.register_handler(
            CoroutineCallback(
                "iq:gateway",
                StanzaPath("iq/gateway"),
                self._handle_gateway_iq,  # type: ignore
            )
        )

    async def search_get_form(self, _gateway_jid, _node, ifrom: JID, iq: Iq):
        """
        Prepare the search form using :attr:`.BaseSession.SEARCH_FIELDS`
        """
        user = user_store.get_by_jid(ifrom)
        if user is None:
            raise XMPPError(text="Search is only allowed for registered users")

        xmpp = self.xmpp

        reply = iq.reply()
        form = reply["search"]["form"]
        form["title"] = xmpp.SEARCH_TITLE
        form["instructions"] = xmpp.SEARCH_INSTRUCTIONS
        for field in xmpp.SEARCH_FIELDS:
            form.append(field.get_xml())
        return reply

    async def _search_query(self, _gateway_jid, _node, ifrom: JID, iq: Iq):
        """
        Handles a search request
        """
        user = user_store.get_by_jid(ifrom)
        if user is None:
            raise XMPPError(text="Search is only allowed for registered users")

        result = await self.xmpp.get_session_from_stanza(iq).on_search(
            iq["search"]["form"].get_values()
        )

        if not result:
            raise XMPPError("item-not-found", text="Nothing was found")

        reply = iq.reply()
        form = reply["search"]["form"]
        for field in result.fields:
            form.add_reported(field.var, label=field.label, type=field.type)
        for item in result.items:
            form.add_item(item)
        return reply

    async def _handle_gateway_iq(self, iq: Iq):
        if iq.get_to() != self.xmpp.boundjid.bare:
            raise XMPPError("bad-request", "This can only be used on the component JID")

        user = user_store.get_by_jid(iq.get_from())
        if user is None:
            raise XMPPError("not-authorized", "Register to the gateway first")

        if len(self.xmpp.SEARCH_FIELDS) > 1:
            raise XMPPError(
                "feature-not-implemented", "Use jabber search for this gateway"
            )

        field = self.xmpp.SEARCH_FIELDS[0]

        reply = iq.reply()
        if iq["type"] == "get":
            reply["gateway"]["desc"] = self.xmpp.SEARCH_TITLE
            reply["gateway"]["prompt"] = field.label
        elif iq["type"] == "set":
            prompt = iq["gateway"]["prompt"]
            session = self.xmpp.session_cls.from_user(user)
            result = await session.on_search({field.var: prompt})
            if result is None or not result.items:
                raise XMPPError(
                    "item-not-found", "No contact was found with the info you provided."
                )
            if len(result.items) > 1:
                raise XMPPError(
                    "bad-request", "Your search yielded more than one result."
                )
            reply["gateway"]["jid"] = result.items[0]["jid"]

        reply.send()



import logging
from functools import cached_property
from types import GenericAlias
from typing import Optional, Union, get_args, get_origin, get_type_hints

import configargparse


class Option:
    DOC_SUFFIX = "__DOC"
    DYNAMIC_DEFAULT_SUFFIX = "__DYNAMIC_DEFAULT"
    SHORT_SUFFIX = "__SHORT"

    def __init__(self, parent: "ConfigModule", name: str):
        self.parent = parent
        self.config_obj = parent.config_obj
        self.name = name

    @cached_property
    def doc(self):
        return getattr(self.config_obj, self.name + self.DOC_SUFFIX)

    @cached_property
    def required(self):
        return not hasattr(
            self.config_obj, self.name + self.DYNAMIC_DEFAULT_SUFFIX
        ) and not hasattr(self.config_obj, self.name)

    @cached_property
    def default(self):
        return getattr(self.config_obj, self.name, None)

    @cached_property
    def short(self):
        return getattr(self.config_obj, self.name + self.SHORT_SUFFIX, None)

    @cached_property
    def nargs(self):
        type_ = get_type_hints(self.config_obj).get(self.name, type(self.default))

        if isinstance(type_, GenericAlias):
            args = get_args(type_)
            if args[1] is Ellipsis:
                return "*"
            else:
                return len(args)

    @cached_property
    def type(self):
        type_ = get_type_hints(self.config_obj).get(self.name, type(self.default))

        if _is_optional(type_):
            type_ = get_args(type_)[0]
        elif isinstance(type_, GenericAlias):
            args = get_args(type_)
            type_ = args[0]

        return type_

    @cached_property
    def names(self):
        res = ["--" + self.name.lower().replace("_", "-")]
        if s := self.short:
            res.append("-" + s)
        return res

    @cached_property
    def kwargs(self):
        kwargs = dict(
            required=self.required,
            help=self.doc,
            env_var=self.name_to_env_var(),
        )
        t = self.type
        if t is bool:
            if self.default:
                kwargs["action"] = "store_false"
            else:
                kwargs["action"] = "store_true"
        else:
            kwargs["type"] = t
            if self.required:
                kwargs["required"] = True
            else:
                kwargs["default"] = self.default
        if n := self.nargs:
            kwargs["nargs"] = n
        return kwargs

    def name_to_env_var(self):
        return self.parent.ENV_VAR_PREFIX + self.name


class ConfigModule:
    ENV_VAR_PREFIX = "SLIDGE_"

    def __init__(
        self, config_obj, parser: Optional[configargparse.ArgumentParser] = None
    ):
        self.config_obj = config_obj
        if parser is None:
            parser = configargparse.ArgumentParser()
        self.parser = parser

        self.add_options_to_parser()

    def _list_options(self):
        return {
            o
            for o in (set(dir(self.config_obj)) | set(get_type_hints(self.config_obj)))
            if o.upper() == o and not o.startswith("_") and "__" not in o
        }

    def set_conf(self, argv: Optional[list[str]] = None):
        if argv is not None:
            # this is ugly, but necessary because for plugin config, we used
            # remaining argv.
            # when using (a) .ini file(s), for bool options, we end-up with
            # remaining pseudo-argv such as --some-bool-opt=true when we really
            # should have just --some-bool-opt
            # TODO: get rid of configargparse and make this cleaner
            options_long = {o.name: o for o in self.options}
            no_explicit_bool = []
            skip_next = False
            for a, aa in zip(argv, argv[1:] + [""]):
                if skip_next:
                    skip_next = False
                    continue
                force_keep = False
                if "=" in a:
                    real_name, _value = a.split("=")
                    opt: Optional[Option] = options_long.get(
                        _argv_to_option_name(real_name)
                    )
                    if opt and opt.type is bool:
                        if opt.default:
                            if _value in _TRUEISH or not _value:
                                continue
                            else:
                                a = real_name
                                force_keep = True
                        else:
                            if _value in _TRUEISH:
                                a = real_name
                                force_keep = True
                            else:
                                continue
                else:
                    upper = _argv_to_option_name(a)
                    opt = options_long.get(upper)
                    if opt and opt.type is bool:
                        if _argv_to_option_name(aa) not in options_long:
                            log.debug("Removing %s from argv", aa)
                            skip_next = True

                if opt:
                    if opt.type is bool:
                        if force_keep or not opt.default:
                            no_explicit_bool.append(a)
                    else:
                        no_explicit_bool.append(a)
                else:
                    no_explicit_bool.append(a)
            log.debug("Removed boolean values from %s to %s", argv, no_explicit_bool)
            argv = no_explicit_bool

        args, rest = self.parser.parse_known_args(argv)
        self.update_dynamic_defaults(args)
        for name in self._list_options():
            value = getattr(args, name.lower())
            log.debug("Setting '%s' to %r", name, value)
            setattr(self.config_obj, name, value)
        return args, rest

    @cached_property
    def options(self) -> list[Option]:
        res = []
        for opt in self._list_options():
            res.append(Option(self, opt))
        return res

    def add_options_to_parser(self):
        p = self.parser
        for o in sorted(self.options, key=lambda x: (not x.required, x.name)):
            p.add_argument(*o.names, **o.kwargs)

    def update_dynamic_defaults(self, args):
        pass


def _is_optional(t):
    if get_origin(t) is Union:
        args = get_args(t)
        if len(args) == 2 and isinstance(None, args[1]):
            return True
    return False


def _argv_to_option_name(arg: str):
    return arg.upper().removeprefix("--").replace("-", "_")


_TRUEISH = {"true", "True", "1", "on", "enabled"}


log = logging.getLogger(__name__)



import hashlib
import io
import logging
from copy import copy
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Type, Union

from PIL import Image, UnidentifiedImageError
from PIL.Image import Image as PILImage
from slixmpp import (
    JID,
    CoroutineCallback,
    Iq,
    Presence,
    StanzaPath,
    register_stanza_plugin,
)
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.base import BasePlugin, register_plugin
from slixmpp.plugins.xep_0060.stanza import Event, EventItem, EventItems, Item
from slixmpp.plugins.xep_0084 import Data as AvatarData
from slixmpp.plugins.xep_0084 import MetaData as AvatarMetadata
from slixmpp.plugins.xep_0172 import UserNick
from slixmpp.plugins.xep_0292.stanza import VCard4
from slixmpp.types import JidStr, OptJidStr

from ..contact.contact import LegacyContact
from ..contact.roster import ContactIsUser
from ..util.db import GatewayUser, user_store
from ..util.sql import db
from ..util.types import AvatarType, LegacyFileIdType, PepItemType
from .cache import CachedAvatar, avatar_cache
from .mixins.lock import NamedLockMixin

if TYPE_CHECKING:
    from slidge import BaseGateway

VCARD4_NAMESPACE = "urn:xmpp:vcard4"


class PepItem:
    @staticmethod
    def from_db(jid: JID, user: Optional[GatewayUser] = None) -> Optional["PepItem"]:
        raise NotImplementedError

    def to_db(self, jid: JID, user: Optional[GatewayUser] = None):
        raise NotImplementedError


class PepAvatar(PepItem):
    def __init__(self, jid: JID):
        self.metadata: Optional[AvatarMetadata] = None
        self.id: Optional[str] = None
        self.jid = jid
        self._avatar_data_path: Optional[Path] = None

    @property
    def data(self) -> Optional[AvatarData]:
        if self._avatar_data_path is None:
            return None
        data = AvatarData()
        data.set_value(self._avatar_data_path.read_bytes())
        return data

    @staticmethod
    def _sha(b: bytes):
        return hashlib.sha1(b).hexdigest()

    def _get_weak_unique_id(self, avatar: AvatarType):
        if isinstance(avatar, str):
            return avatar  # url
        elif isinstance(avatar, bytes):
            return self._sha(avatar)
        elif isinstance(avatar, Path):
            return self._sha(avatar.read_bytes())

    @staticmethod
    async def _get_image(avatar: AvatarType) -> PILImage:
        if isinstance(avatar, str):
            # async with aiohttp.ClientSession() as session:
            async with avatar_cache.http.get(avatar) as response:
                return Image.open(io.BytesIO(await response.read()))
        elif isinstance(avatar, bytes):
            return Image.open(io.BytesIO(avatar))
        elif isinstance(avatar, Path):
            return Image.open(avatar)
        else:
            raise TypeError("Avatar must be bytes, a Path or a str (URL)", avatar)

    async def set_avatar(
        self, avatar: AvatarType, unique_id: Optional[LegacyFileIdType] = None
    ):
        if unique_id is None:
            if isinstance(avatar, str):
                await self._set_avatar_from_url_alone(avatar)
                return
            unique_id = self._get_weak_unique_id(avatar)

        await self._set_avatar_from_unique_id(avatar, unique_id)

    async def _set_avatar_from_unique_id(
        self, avatar: AvatarType, unique_id: LegacyFileIdType
    ):
        cached_avatar = avatar_cache.get(unique_id)
        if cached_avatar:
            # this shouldn't be necessary but here to re-use avatars downloaded
            # before the change introducing the JID to unique ID mapping
            avatar_cache.store_jid(self.jid, unique_id)
        else:
            img = await self._get_image(avatar)
            cached_avatar = await avatar_cache.convert_and_store(
                img, unique_id, self.jid
            )

        self._set_avatar_from_cache(cached_avatar)

    async def _set_avatar_from_url_alone(self, url: str):
        cached_avatar = await avatar_cache.get_avatar_from_url_alone(url, self.jid)
        self._set_avatar_from_cache(cached_avatar)

    def _set_avatar_from_cache(self, cached_avatar: CachedAvatar):
        metadata = AvatarMetadata()
        self.id = cached_avatar.hash
        metadata.add_info(
            id=cached_avatar.hash,
            itype="image/png",
            ibytes=cached_avatar.path.stat().st_size,
            height=str(cached_avatar.height),
            width=str(cached_avatar.width),
        )
        self.metadata = metadata
        self._avatar_data_path = cached_avatar.path

    @staticmethod
    def from_db(jid: JID, user: Optional[GatewayUser] = None) -> Optional["PepAvatar"]:
        cached_id = db.avatar_get(jid)
        if cached_id is None:
            return None
        item = PepAvatar(jid)
        cached_avatar = avatar_cache.get(cached_id)
        if cached_avatar is None:
            raise XMPPError("internal-server-error")
        item._set_avatar_from_cache(cached_avatar)
        return item

    def to_db(self, jid: JID, user=None):
        cached_id = avatar_cache.get_cached_id_for(jid)
        if cached_id is None:
            log.warning("Could not store avatar for %s", jid)
            return
        db.avatar_store(jid, cached_id)

    @staticmethod
    def remove_from_db(jid: JID):
        db.avatar_delete(jid)


class PepNick(PepItem):
    def __init__(self, nick: Optional[str] = None):
        nickname = UserNick()
        if nick is not None:
            nickname["nick"] = nick
        self.nick = nickname
        self.__nick_str = nick

    @staticmethod
    def from_db(jid: JID, user: Optional[GatewayUser] = None) -> Optional["PepNick"]:
        if user is None:
            raise XMPPError("not-allowed")
        nick = db.nick_get(jid, user)
        if nick is None:
            return None
        return PepNick(nick)

    def to_db(self, jid: JID, user: Optional[GatewayUser] = None):
        if user is None:
            raise XMPPError("not-allowed")
        db.nick_store(jid, str(self.__nick_str), user)


class PubSubComponent(NamedLockMixin, BasePlugin):
    xmpp: "BaseGateway"

    name = "pubsub"
    description = "Pubsub component"
    dependencies = {
        "xep_0030",
        "xep_0060",
        # "xep_0084",
        "xep_0115",
        "xep_0163",
    }
    default_config = {"component_name": None}
    component_name: str

    def __init__(self, *a, **kw):
        super(PubSubComponent, self).__init__(*a, **kw)
        register_stanza_plugin(EventItem, UserNick)

    def plugin_init(self):
        self.xmpp.register_handler(
            CoroutineCallback(
                "pubsub_get_avatar_data",
                StanzaPath(f"iq@type=get/pubsub/items@node={AvatarData.namespace}"),
                self._get_avatar_data,  # type:ignore
            )
        )
        self.xmpp.register_handler(
            CoroutineCallback(
                "pubsub_get_avatar_metadata",
                StanzaPath(f"iq@type=get/pubsub/items@node={AvatarMetadata.namespace}"),
                self._get_avatar_metadata,  # type:ignore
            )
        )
        self.xmpp.register_handler(
            CoroutineCallback(
                "pubsub_get_vcard",
                StanzaPath(f"iq@type=get/pubsub/items@node={VCARD4_NAMESPACE}"),
                self._get_vcard,  # type:ignore
            )
        )
        self.xmpp.add_event_handler("presence_available", self._on_presence_available)

        disco = self.xmpp.plugin["xep_0030"]
        disco.add_identity("pubsub", "pep", self.component_name)
        disco.add_identity("account", "registered", self.component_name)
        disco.add_feature("http://jabber.org/protocol/pubsub#event")
        disco.add_feature("http://jabber.org/protocol/pubsub#retrieve-items")
        disco.add_feature("http://jabber.org/protocol/pubsub#persistent-items")

    async def _on_presence_available(self, p: Presence):
        if p.get_plugin("muc_join", check=True) is not None:
            log.debug("Ignoring MUC presence here")
            return

        from_ = p.get_from()
        ver_string = p["caps"]["ver"]
        info = None

        to = p.get_to()

        # we don't want to push anything for contacts that are not in the user's roster
        if to != self.xmpp.boundjid.bare:
            session = self.xmpp.get_session_from_stanza(p)

            if session is None:
                return

            await session.contacts.ready
            try:
                contact = await session.contacts.by_jid(to)
            except XMPPError as e:
                log.debug(
                    "Could not determine if %s was added to the roster: %s", to, e
                )
                return
            except Exception as e:
                log.warning("Could not determine if %s was added to the roster.", to)
                log.exception(e)
                return
            if not contact.is_friend:
                return

        if ver_string:
            info = await self.xmpp.plugin["xep_0115"].get_caps(from_)
        if info is None:
            async with self.lock(from_):
                iq = await self.xmpp.plugin["xep_0030"].get_info(from_)
            info = iq["disco_info"]
        features = info["features"]
        if AvatarMetadata.namespace + "+notify" in features:
            try:
                pep_avatar = await self._get_authorized_avatar(p)
            except XMPPError:
                pass
            else:
                if pep_avatar.metadata is None:
                    raise XMPPError("internal-server-error", "Avatar but no metadata?")
                await self._broadcast(
                    data=pep_avatar.metadata,
                    from_=p.get_to(),
                    to=from_,
                    id=pep_avatar.metadata["info"]["id"],
                )
        if UserNick.namespace + "+notify" in features:
            try:
                pep_nick = await self._get_authorized_nick(p)
            except XMPPError:
                pass
            else:
                await self._broadcast(data=pep_nick.nick, from_=p.get_to(), to=from_)

        if VCARD4_NAMESPACE + "+notify" in features:
            await self.broadcast_vcard_event(p.get_to(), to=from_)

    async def broadcast_vcard_event(self, from_, to):
        item = Item()
        item.namespace = VCARD4_NAMESPACE
        item["id"] = "current"
        vcard: VCard4 = await self.xmpp["xep_0292_provider"].get_vcard(from_, to)
        # The vcard content should NOT be in this event according to the spec:
        # https://xmpp.org/extensions/xep-0292.html#sect-idm45669698174224
        # but movim expects it to be here, and I guess

        log.debug("Broadcast vcard4 event: %s", vcard)
        await self._broadcast(
            data=vcard,
            from_=JID(from_).bare,
            to=to,
            id="current",
            node=VCARD4_NAMESPACE,
        )

    async def _get_authorized_item(
        self, cls: Type[PepItemType], stanza: Union[Iq, Presence]
    ) -> PepItemType:
        sto = stanza.get_to()
        user = user_store.get_by_jid(stanza.get_from())
        item = cls.from_db(sto, user)
        if item is None:
            raise XMPPError("item-not-found")

        if sto != self.xmpp.boundjid.bare:
            session = self.xmpp.get_session_from_stanza(stanza)
            await session.contacts.by_jid(sto)

        return item  # type:ignore

    async def _get_authorized_avatar(self, stanza: Union[Iq, Presence]) -> PepAvatar:
        return await self._get_authorized_item(PepAvatar, stanza)

    async def _get_authorized_nick(self, stanza: Union[Iq, Presence]) -> PepNick:
        return await self._get_authorized_item(PepNick, stanza)

    async def _get_avatar_data(self, iq: Iq):
        pep_avatar = await self._get_authorized_avatar(iq)

        requested_items = iq["pubsub"]["items"]
        if len(requested_items) == 0:
            self._reply_with_payload(iq, pep_avatar.data, pep_avatar.id)
        else:
            for item in requested_items:
                if item["id"] == pep_avatar.id:
                    self._reply_with_payload(iq, pep_avatar.data, pep_avatar.id)
                    return
            else:
                raise XMPPError("item-not-found")

    async def _get_avatar_metadata(self, iq: Iq):
        pep_avatar = await self._get_authorized_avatar(iq)

        requested_items = iq["pubsub"]["items"]
        if len(requested_items) == 0:
            self._reply_with_payload(iq, pep_avatar.metadata, pep_avatar.id)
        else:
            for item in requested_items:
                if item["id"] == pep_avatar.id:
                    self._reply_with_payload(iq, pep_avatar.metadata, pep_avatar.id)
                    return
            else:
                raise XMPPError("item-not-found")

    async def _get_vcard(self, iq: Iq):
        # this is not the proper way that clients should retrieve VCards, but
        # gajim does it this way.
        # https://xmpp.org/extensions/xep-0292.html#sect-idm45669698174224
        vcard: VCard4 = await self.xmpp["xep_0292_provider"].get_vcard(
            iq.get_to().bare, iq.get_from().bare
        )
        log.debug("VCARD: %s -- %s -- %s", iq.get_to().bare, iq.get_from().bare, vcard)
        if vcard is None:
            raise XMPPError("item-not-found")
        self._reply_with_payload(iq, vcard, "current", VCARD4_NAMESPACE)

    @staticmethod
    def get_avatar(jid: JID):
        return PepAvatar.from_db(jid)

    @staticmethod
    def _reply_with_payload(
        iq: Iq,
        payload: Optional[Union[AvatarMetadata, AvatarData, VCard4]],
        id_: Optional[str],
        namespace: Optional[str] = None,
    ):
        result = iq.reply()
        item = Item()
        if payload:
            item.set_payload(payload.xml)
            item["id"] = id_
            result["pubsub"]["items"]["node"] = (
                namespace if namespace else payload.namespace
            )
        result["pubsub"]["items"].append(item)
        result.send()

    async def _broadcast(self, data, from_: JidStr, to: OptJidStr = None, **kwargs):
        from_ = JID(from_)
        if from_ != self.xmpp.boundjid.bare and to is not None:
            to = JID(to)
            session = self.xmpp.get_session_from_jid(to)
            if session is None:
                return
            await session.ready
            try:
                entity = await session.get_contact_or_group_or_participant(from_)
            except ContactIsUser:
                return
            if isinstance(entity, LegacyContact) and not entity.is_friend:
                return

        item = EventItem()
        if data:
            item.set_payload(data.xml)
        for k, v in kwargs.items():
            item[k] = v

        items = EventItems()
        items.append(item)
        items["node"] = kwargs.get("node") or data.namespace

        event = Event()
        event.append(items)

        msg = self.xmpp.Message()
        msg.set_type("headline")
        msg.set_from(from_)
        msg.append(event)

        if to is None:
            for u in user_store.get_all():
                new_msg = copy(msg)
                new_msg.set_to(u.bare_jid)
                new_msg.send()
        else:
            msg.set_to(to)
            msg.send()

    async def set_avatar(
        self,
        jid: JidStr,
        avatar: Optional[AvatarType] = None,
        broadcast_to: OptJidStr = None,
        unique_id=None,
        broadcast=True,
    ):
        jid = JID(jid)
        if avatar is None:
            PepAvatar.remove_from_db(jid)
            await self._broadcast(AvatarMetadata(), jid, broadcast_to)
            avatar_cache.delete_jid(jid)
        else:
            pep_avatar = PepAvatar(jid)
            try:
                await pep_avatar.set_avatar(avatar, unique_id)
            except (UnidentifiedImageError, FileNotFoundError) as e:
                log.warning("Failed to set avatar for %s: %r", self, e)
                return
            pep_avatar.to_db(jid)
            if pep_avatar.metadata is None:
                raise RuntimeError
            if not broadcast:
                return
            await self._broadcast(
                pep_avatar.metadata,
                jid,
                broadcast_to,
                id=pep_avatar.metadata["info"]["id"],
            )

    async def set_avatar_from_cache(
        self, jid: JID, send_empty: bool, broadcast_to: OptJidStr = None, broadcast=True
    ):
        uid = avatar_cache.get_cached_id_for(jid)
        if uid is None:
            if not send_empty:
                return
            self.xmpp.loop.create_task(
                self.set_avatar(jid, None, broadcast_to, uid, broadcast)
            )
            return
        cached_avatar = avatar_cache.get(str(uid))
        if cached_avatar is None:
            # should not happen but well…
            log.warning(
                "Something is wrong with the avatar, %s won't have an "
                "avatar because avatar not found in cache",
                jid,
            )
            return
        self.xmpp.loop.create_task(
            self.set_avatar(jid, cached_avatar.path, broadcast_to, uid, broadcast)
        )

    def set_nick(
        self,
        user: GatewayUser,
        jid: JidStr,
        nick: Optional[str] = None,
    ):
        jid = JID(jid)
        nickname = PepNick(nick)
        nickname.to_db(jid, user)
        log.debug("New nickname: %s", nickname.nick)
        self.xmpp.loop.create_task(self._broadcast(nickname.nick, jid, user.bare_jid))

    async def broadcast_all(self, from_: JID, to: JID):
        """
        Force push avatar and nick for a stored JID.
        """
        a = PepAvatar.from_db(from_)
        if a:
            if a.metadata:
                await self._broadcast(
                    a.metadata, from_, to, id=a.metadata["info"]["id"]
                )
            else:
                log.warning("No metadata associated to this cached avatar?!")
        n = PepNick.from_db(from_, user_store.get_by_jid(to))
        if n:
            await self._broadcast(n.nick, from_, to)


log = logging.getLogger(__name__)
register_plugin(PubSubComponent)



from slixmpp import Iq
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0004 import Form
from slixmpp.types import JidStr

from . import stanza


class XEP_0490(BasePlugin):
    """
    XEP-0490: Message Displayed Synchronization
    """

    name = "xep_0490"
    description = "XEP-0490: Message Displayed Synchronization"
    dependencies = {"xep_0060", "xep_0163", "xep_0359"}
    stanza = stanza

    def plugin_init(self):
        stanza.register_plugin()
        self.xmpp.plugin["xep_0163"].register_pep(
            "message_displayed_synchronization",
            stanza.Displayed,
        )

    def flag_chat(self, chat: JidStr, stanza_id: str, **kwargs) -> Iq:
        displayed = stanza.Displayed()
        displayed["stanza_id"]["id"] = stanza_id
        return self.xmpp.plugin["xep_0163"].publish(
            displayed, node=stanza.NS, options=PUBLISH_OPTIONS, id=str(chat), **kwargs
        )

    def catch_up(self, **kwargs):
        return self.xmpp.plugin["xep_0060"].get_items(
            self.xmpp.boundjid.bare, stanza.NS, **kwargs
        )


PUBLISH_OPTIONS = Form()
PUBLISH_OPTIONS["type"] = "submit"
PUBLISH_OPTIONS.add_field(
    "FORM_TYPE", "hidden", value="http://jabber.org/protocol/pubsub#publish-options"
)
PUBLISH_OPTIONS.add_field("pubsub#persist_items", value="true")
PUBLISH_OPTIONS.add_field("pubsub#max_items", value="max")
PUBLISH_OPTIONS.add_field("pubsub#send_last_published_item", value="never")
PUBLISH_OPTIONS.add_field("pubsub#access_model", value="whitelist")



from abc import ABCMeta
from typing import TYPE_CHECKING

from slixmpp import JID

from ...util.types import MessageOrPresenceTypeVar

if TYPE_CHECKING:
    from slidge.core.gateway import BaseGateway
    from slidge.core.session import BaseSession
    from slidge.util.db import GatewayUser


class MetaBase(ABCMeta):
    pass


class Base:
    session: "BaseSession" = NotImplemented
    xmpp: "BaseGateway" = NotImplemented
    user: "GatewayUser" = NotImplemented

    jid: JID = NotImplemented
    name: str = NotImplemented


class BaseSender(Base):
    def _send(
        self, stanza: MessageOrPresenceTypeVar, **send_kwargs
    ) -> MessageOrPresenceTypeVar:
        raise NotImplementedError



from typing import Optional

from slixmpp import register_stanza_plugin
from slixmpp.plugins.xep_0234.stanza import File
from slixmpp.xmlstream import ElementBase

NS = "urn:xmpp:thumbs:1"


class Thumbnail(ElementBase):
    name = plugin_attrib = "thumbnail"
    namespace = NS
    interfaces = {"uri", "media-type", "width", "height"}

    def get_width(self) -> float:
        return _int_or_none(self._get_attr("width"))

    def get_height(self) -> float:
        return _int_or_none(self._get_attr("height"))

    def set_width(self, v: int) -> None:
        self._set_attr("width", str(v))

    def set_height(self, v: int) -> None:
        self._set_attr("height", str(v))


def _int_or_none(v) -> Optional[int]:
    try:
        return int(v)
    except ValueError:
        return None


def register_plugin():
    register_stanza_plugin(File, Thumbnail)



import asyncio
import logging
from typing import TYPE_CHECKING, Generic, Optional, Type

from slixmpp import JID
from slixmpp.jid import JID_UNESCAPE_TRANSFORMATIONS, _unescape_node

from ..core.mixins.lock import NamedLockMixin
from ..util import SubclassableOnce
from ..util.types import LegacyContactType, LegacyUserIdType
from .contact import LegacyContact

if TYPE_CHECKING:
    from ..core.session import BaseSession


class ContactIsUser(Exception):
    pass


class LegacyRoster(
    Generic[LegacyUserIdType, LegacyContactType],
    NamedLockMixin,
    metaclass=SubclassableOnce,
):
    """
    Virtual roster of a gateway user, that allows to represent all
    of their contacts as singleton instances (if used properly and not too bugged).

    Every :class:`.BaseSession` instance will have its own :class:`.LegacyRoster` instance
    accessible via the :attr:`.BaseSession.contacts` attribute.

    Typically, you will mostly use the :meth:`.LegacyRoster.by_legacy_id` function to
    retrieve a contact instance.

    You might need to override :meth:`.LegacyRoster.legacy_id_to_jid_username` and/or
    :meth:`.LegacyRoster.jid_username_to_legacy_id` to incorporate some custom logic
    if you need some characters when translation JID user parts and legacy IDs.
    """

    def __init__(self, session: "BaseSession"):
        self._contact_cls: Type[LegacyContactType] = (
            LegacyContact.get_self_or_unique_subclass()
        )
        self._contact_cls.xmpp = session.xmpp

        self.session = session
        self._contacts_by_bare_jid: dict[str, LegacyContactType] = {}
        self._contacts_by_legacy_id: dict[LegacyUserIdType, LegacyContactType] = {}
        self.log = logging.getLogger(f"{self.session.user.bare_jid}:roster")
        self.user_legacy_id: Optional[LegacyUserIdType] = None
        self.ready: asyncio.Future[bool] = self.session.xmpp.loop.create_future()
        super().__init__()

    def __repr__(self):
        return f"<Roster of {self.session.user}>"

    def __iter__(self):
        return iter(self._contacts_by_legacy_id.values())

    async def __finish_init_contact(
        self, legacy_id: LegacyUserIdType, jid_username: str, *args, **kwargs
    ):
        c = self._contact_cls(self.session, legacy_id, jid_username, *args, **kwargs)
        async with self.lock(("finish", c)):
            if legacy_id in self._contacts_by_legacy_id:
                self.log.debug("Already updated %s", c)
                return c
            await c.avatar_wrap_update_info()
            self._contacts_by_legacy_id[legacy_id] = c
            self._contacts_by_bare_jid[c.jid.bare] = c
        return c

    def known_contacts(self, only_friends=True) -> dict[str, LegacyContactType]:
        if only_friends:
            return {j: c for j, c in self._contacts_by_bare_jid.items() if c.is_friend}
        return self._contacts_by_bare_jid

    async def by_jid(self, contact_jid: JID) -> LegacyContactType:
        # """
        # Retrieve a contact by their JID
        #
        # If the contact was not instantiated before, it will be created
        # using :meth:`slidge.LegacyRoster.jid_username_to_legacy_id` to infer their
        # legacy user ID.
        #
        # :param contact_jid:
        # :return:
        # """
        username = contact_jid.node
        async with self.lock(("username", username)):
            bare = contact_jid.bare
            c = self._contacts_by_bare_jid.get(bare)
            if c is None:
                legacy_id = await self.jid_username_to_legacy_id(username)
                log.debug("Contact %s not found", contact_jid)
                if self.get_lock(("legacy_id", legacy_id)):
                    log.debug("Already updating %s", contact_jid)
                    return await self.by_legacy_id(legacy_id)
                c = await self.__finish_init_contact(legacy_id, username)
            return c

    async def by_legacy_id(
        self, legacy_id: LegacyUserIdType, *args, **kwargs
    ) -> LegacyContactType:
        """
        Retrieve a contact by their legacy_id

        If the contact was not instantiated before, it will be created
        using :meth:`slidge.LegacyRoster.legacy_id_to_jid_username` to infer their
        legacy user ID.

        :param legacy_id:
        :param args: arbitrary additional positional arguments passed to the contact constructor.
            Requires subclassing LegacyContact.__init__ to accept those.
            This is useful for networks where you fetch the contact list and information
            about these contacts in a single request
        :param kwargs: arbitrary keyword arguments passed to the contact constructor
        :return:
        """
        if legacy_id == self.user_legacy_id:
            raise ContactIsUser
        async with self.lock(("legacy_id", legacy_id)):
            c = self._contacts_by_legacy_id.get(legacy_id)
            if c is None:
                username = await self.legacy_id_to_jid_username(legacy_id)
                log.debug("Contact %s not found", legacy_id)
                if self.get_lock(("username", username)):
                    log.debug("Already updating %s", username)
                    jid = JID()
                    jid.node = username
                    jid.domain = self.session.xmpp.boundjid.bare
                    return await self.by_jid(jid)
                c = await self.__finish_init_contact(
                    legacy_id, username, *args, **kwargs
                )
            return c

    async def by_stanza(self, s) -> LegacyContact:
        # """
        # Retrieve a contact by the destination of a stanza
        #
        # See :meth:`slidge.Roster.by_legacy_id` for more info.
        #
        # :param s:
        # :return:
        # """
        return await self.by_jid(s.get_to())

    async def legacy_id_to_jid_username(self, legacy_id: LegacyUserIdType) -> str:
        """
        Convert a legacy ID to a valid 'user' part of a JID

        Should be overridden for cases where the str conversion of
        the legacy_id is not enough, e.g., if it is case-sensitive or contains
        forbidden characters not covered by :xep:`0106`.

        :param legacy_id:
        """
        return str(legacy_id).translate(ESCAPE_TABLE)

    async def jid_username_to_legacy_id(self, jid_username: str) -> LegacyUserIdType:
        """
        Convert a JID user part to a legacy ID.

        Should be overridden in case legacy IDs are not strings, or more generally
        for any case where the username part of a JID (unescaped with to the mapping
        defined by :xep:`0106`) is not enough to identify a contact on the legacy network.

        Default implementation is an identity operation

        :param jid_username: User part of a JID, ie "user" in "user@example.com"
        :return: An identifier for the user on the legacy network.
        """
        return _unescape_node(jid_username)

    async def fill(self):
        """
        Populate slidge's "virtual roster".

        Override this and in it, ``await self.by_legacy_id(contact_id)``
        for the every legacy contacts of the user for which you'd like to
        set an avatar, nickname, vcard…

        Await ``Contact.add_to_roster()`` in here to add the contact to the
        user's XMPP roster.
        """
        pass


ESCAPE_TABLE = "".maketrans({v: k for k, v in JID_UNESCAPE_TRANSFORMATIONS.items()})
log = logging.getLogger(__name__)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.plugins.xep_0359 import OriginID
from slixmpp.stanza import Message
from slixmpp.xmlstream import ElementBase, register_stanza_plugin

NS = "urn:xmpp:message-retract:1"


class Retract(ElementBase):
    namespace = NS
    name = "retract"
    plugin_attrib = "retract"


class Retracted(ElementBase):
    namespace = NS
    name = "retracted"
    plugin_attrib = "retracted"
    interfaces = {"stamp"}


def register_plugins():
    register_stanza_plugin(Message, Retract)
    register_stanza_plugin(Message, Retracted)
    register_stanza_plugin(Retracted, OriginID)



"""
Mixins
"""

from .avatar import AvatarMixin
from .disco import ChatterDiscoMixin
from .message import MessageCarbonMixin, MessageMixin
from .presence import PresenceMixin


class FullMixin(ChatterDiscoMixin, MessageMixin, PresenceMixin):
    pass


class FullCarbonMixin(ChatterDiscoMixin, MessageCarbonMixin, PresenceMixin):
    pass


__all__ = ("AvatarMixin",)



from copy import copy
from datetime import datetime, timezone
from typing import Optional, Union
from xml.etree import ElementTree as ET

from slixmpp import Message
from slixmpp.plugins.xep_0297 import Forwarded


def fix_namespaces(xml, old="{jabber:component:accept}", new="{jabber:client}"):
    """
    Hack to fix namespaces between jabber:component and jabber:client

    Acts in-place.

    :param xml:
    :param old:
    :param new:
    """
    xml.tag = xml.tag.replace(old, new)
    for child in xml:
        fix_namespaces(child, old, new)


class HistoryMessage:
    def __init__(self, stanza: Union[Message, str], when: Optional[datetime] = None):
        if isinstance(stanza, str):
            from_db = True
            stanza = Message(xml=ET.fromstring(stanza))
        else:
            from_db = False

        self.id = stanza["stanza_id"]["id"]
        self.when: datetime = (
            when or stanza["delay"]["stamp"] or datetime.now(tz=timezone.utc)
        )

        if not from_db:
            del stanza["delay"]
            del stanza["markable"]
            del stanza["hint"]
            del stanza["chat_state"]
            if not stanza["body"]:
                del stanza["body"]
            fix_namespaces(stanza.xml)

        self.stanza: Message = stanza

    @property
    def stanza_component_ns(self):
        stanza = copy(self.stanza)
        fix_namespaces(
            stanza.xml, old="{jabber:client}", new="{jabber:component:accept}"
        )
        return stanza

    def forwarded(self):
        forwarded = Forwarded()
        forwarded["delay"]["stamp"] = self.when
        forwarded.append(self.stanza)
        return forwarded



import logging
from typing import TYPE_CHECKING, NamedTuple, Optional

from slixmpp import JID, CoroutineCallback, Iq, StanzaPath
from slixmpp.plugins.base import BasePlugin, register_plugin
from slixmpp.plugins.xep_0292.stanza import NS, VCard4
from slixmpp.types import JidStr

from slidge.contact import LegacyContact

if TYPE_CHECKING:
    from slidge.core.gateway import BaseGateway


class StoredVCard(NamedTuple):
    content: VCard4
    authorized_jids: set[JidStr]


class VCard4Provider(BasePlugin):
    xmpp: "BaseGateway"

    name = "xep_0292_provider"
    description = "VCard4 Provider"
    dependencies = {"xep_0030"}

    def __init__(self, *a, **k):
        super(VCard4Provider, self).__init__(*a, **k)
        self._vcards = dict[JidStr, StoredVCard]()

    def plugin_init(self):
        self.xmpp.register_handler(
            CoroutineCallback(
                "get_vcard",
                StanzaPath(f"iq@type=get/vcard"),
                self.handle_vcard_get,  # type:ignore
            )
        )

        self.xmpp.plugin["xep_0030"].add_feature(NS)

    def _get_cached_vcard(self, jid: JidStr, requested_by: JidStr) -> Optional[VCard4]:
        vcard = self._vcards.get(JID(jid).bare)
        if vcard:
            if auth := vcard.authorized_jids:
                if JID(requested_by).bare in auth:
                    return vcard.content
            else:
                return vcard.content
        return None

    async def get_vcard(self, jid: JidStr, requested_by: JidStr) -> Optional[VCard4]:
        if vcard := self._get_cached_vcard(jid, requested_by):
            log.debug("Found a cached vcard")
            return vcard
        if not hasattr(self.xmpp, "get_session_from_jid"):
            return None
        jid = JID(jid)
        requested_by = JID(requested_by)
        session = self.xmpp.get_session_from_jid(requested_by)
        if session is None:
            return
        entity = await session.get_contact_or_group_or_participant(jid)
        if isinstance(entity, LegacyContact):
            log.debug("Fetching vcard")
            await entity.fetch_vcard()
            return self._get_cached_vcard(jid, requested_by)
        return None

    async def handle_vcard_get(self, iq: Iq):
        r = iq.reply()
        if vcard := await self.get_vcard(iq.get_to().bare, iq.get_from().bare):
            r.append(vcard)
        else:
            r.enable("vcard")
        r.send()

    def set_vcard(
        self,
        jid: JidStr,
        vcard: VCard4,
        /,
        authorized_jids: Optional[set[JidStr]] = None,
    ):
        cache = self._vcards.get(jid)
        new = StoredVCard(
            vcard, authorized_jids if authorized_jids is not None else set()
        )
        self._vcards[jid] = new
        if cache == new:
            return
        if self.xmpp["pubsub"] and authorized_jids:
            for to in authorized_jids:
                self.xmpp.loop.create_task(
                    self.xmpp["pubsub"].broadcast_vcard_event(jid, to)
                )


register_plugin(VCard4Provider)
log = logging.getLogger(__name__)



"""
This module covers a backend for storing user data persistently and managing a
pseudo-roster for the gateway component.
"""

import dataclasses
import datetime
import logging
import os.path
import shelve
from io import BytesIO
from os import PathLike
from typing import Iterable, Optional, Union

from pickle_secure import Pickler, Unpickler
from slixmpp import JID, Iq, Message, Presence

from .sql import db


# noinspection PyUnresolvedReferences
class EncryptedShelf(shelve.DbfilenameShelf):
    cache: dict
    dict: dict
    writeback: bool
    keyencoding: str
    _protocol: int

    def __init__(
        self, filename: PathLike, key: str, flag="c", protocol=None, writeback=False
    ):
        super().__init__(str(filename), flag, protocol, writeback)
        self.secret_key = key

    def __getitem__(self, key):
        try:
            value = self.cache[key]
        except KeyError:
            f = BytesIO(self.dict[key.encode(self.keyencoding)])
            value = Unpickler(f, key=self.secret_key).load()  # type:ignore
            if self.writeback:
                self.cache[key] = value
        return value

    def __setitem__(self, key, value):
        if self.writeback:
            self.cache[key] = value
        f = BytesIO()
        p = Pickler(f, self._protocol, key=self.secret_key)  # type:ignore
        p.dump(value)
        self.dict[key.encode(self.keyencoding)] = f.getvalue()


@dataclasses.dataclass
class GatewayUser:
    """
    A gateway user
    """

    bare_jid: str
    """Bare JID of the user"""
    registration_form: dict[str, Optional[str]]
    """Content of the registration form, as a dict"""
    plugin_data: Optional[dict] = None
    registration_date: Optional[datetime.datetime] = None

    def __hash__(self):
        return hash(self.bare_jid)

    def __repr__(self):
        return f"<User {self.bare_jid}>"

    def __post_init__(self):
        if self.registration_date is None:
            self.registration_date = datetime.datetime.now()

    @property
    def jid(self) -> JID:
        """
        The user's (bare) JID

        :return:
        """
        return JID(self.bare_jid)

    def get(self, field: str, default: str = "") -> Optional[str]:
        # """
        # Get fields from the registration form (required to comply with slixmpp backend protocol)
        #
        # :param field: Name of the field
        # :param default: Default value to return if the field is not present
        #
        # :return: Value of the field
        # """
        return self.registration_form.get(field, default)

    def commit(self):
        db.user_store(self)
        user_store.commit(self)


class UserStore:
    """
    Basic user store implementation using shelve from the python standard library

    Set_file must be called before it is usable
    """

    def __init__(self):
        self._users: shelve.Shelf[GatewayUser] = None  # type: ignore

    def set_file(self, filename: PathLike, secret_key: Optional[str] = None):
        """
        Set the file to use to store user data

        :param filename: Path to the shelf file
        :param secret_key: Secret key to store files encrypted on disk
        """
        if self._users is not None:
            raise RuntimeError("Shelf file already set!")
        if os.path.exists(filename):
            log.info("Using existing slidge DB: %s", filename)
        else:
            log.info("Creating a new slidge DB: %s", filename)
        if secret_key:
            self._users = EncryptedShelf(filename, key=secret_key)
        else:
            self._users = shelve.open(str(filename))
        for user in self._users.values():
            db.user_store(user)
        log.info("Registered users in the DB: %s", list(self._users.keys()))

    def get_all(self) -> Iterable[GatewayUser]:
        """
        Get all users in the store

        :return: An iterable of GatewayUsers
        """
        return self._users.values()

    def add(self, jid: JID, registration_form: dict[str, Optional[str]]):
        """
        Add a user to the store.

        NB: there is no reason to call this manually, as this should be covered
        by slixmpp XEP-0077 and XEP-0100 plugins

        :param jid: JID of the gateway user
        :param registration_form: Content of the registration form (:xep:`0077`)
        """
        log.debug("Adding user %s", jid)
        self._users[jid.bare] = user = GatewayUser(
            bare_jid=jid.bare,
            registration_form=registration_form,
            registration_date=datetime.datetime.now(),
        )
        self._users.sync()
        user.commit()
        log.debug("Store: %s", self._users)

    def commit(self, user: GatewayUser):
        self._users[user.bare_jid] = user
        self._users.sync()

    def get(self, _gateway_jid, _node, ifrom: JID, iq) -> Optional[GatewayUser]:
        """
        Get a user from the store

        NB: there is no reason to call this, it is used by SliXMPP internal API

        :param _gateway_jid:
        :param _node:
        :param ifrom:
        :param iq:
        :return:
        """
        if ifrom is None:  # bug in SliXMPP's XEP_0100 plugin
            ifrom = iq["from"]
        log.debug("Getting user %s", ifrom.bare)
        return self._users.get(ifrom.bare)

    def remove(self, _gateway_jid, _node, ifrom: JID, _iq):
        """
        Remove a user from the store

        NB: there is no reason to call this, it is used by SliXMPP internal API
        """
        self.remove_by_jid(ifrom)

    def remove_by_jid(self, jid: JID):
        """
        Remove a user from the store, by JID
        """
        j = jid.bare
        log.debug("Removing user %s", j)
        db.user_del(self._users[j])
        del self._users[j]
        self._users.sync()

    def get_by_jid(self, jid: JID) -> Optional[GatewayUser]:
        """
        Convenience function to get a user from their JID.

        :param jid: JID of the gateway user
        :return:
        """
        return self._users.get(jid.bare)

    def get_by_stanza(self, s: Union[Presence, Message, Iq]) -> Optional[GatewayUser]:
        """
        Convenience function to get a user from a stanza they sent.

        :param s: A stanza sent by the gateway user
        :return:
        """
        return self.get_by_jid(s.get_from())

    def close(self):
        self._users.sync()
        self._users.close()


user_store = UserStore()
"""
A persistent store for slidge users. Not public, but I didn't find how to hide
it from the docs!
"""

log = logging.getLogger(__name__)



import logging
import string
import stringprep
import uuid
import warnings
from copy import copy
from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING, Optional, Union

from slixmpp import JID, InvalidJID, Message, Presence
from slixmpp.plugins.xep_0045.stanza import MUCAdminItem
from slixmpp.stringprep import StringprepError, resourceprep
from slixmpp.types import MessageTypes, OptJid
from slixmpp.util.stringprep_profiles import StringPrepError, prohibit_output

from ..contact import LegacyContact
from ..core.mixins import ChatterDiscoMixin, MessageMixin, PresenceMixin
from ..util import SubclassableOnce, strip_illegal_chars
from ..util.sql import CachedPresence
from ..util.types import (
    Hat,
    LegacyMessageType,
    MessageOrPresenceTypeVar,
    MucAffiliation,
    MucRole,
)

if TYPE_CHECKING:
    from .room import LegacyMUC


def strip_non_printable(nickname: str):
    new = (
        "".join(x for x in nickname if x in string.printable)
        + f"-slidge-{hash(nickname)}"
    )
    warnings.warn(f"Could not use {nickname} as a nickname, using {new}")
    return new


class LegacyParticipant(
    PresenceMixin,
    MessageMixin,
    ChatterDiscoMixin,
    metaclass=SubclassableOnce,
):
    """
    A legacy participant of a legacy group chat.
    """

    mtype: MessageTypes = "groupchat"
    _can_send_carbon = False
    USE_STANZA_ID = True
    STRIP_SHORT_DELAY = False

    def __init__(
        self,
        muc: "LegacyMUC",
        nickname: Optional[str] = None,
        is_user=False,
        is_system=False,
    ):
        super().__init__()
        self._hats = list[Hat]()
        self.muc = muc
        self.session = session = muc.session
        self.user = session.user
        self.xmpp = session.xmpp
        self._role: MucRole = "participant"
        self._affiliation: MucAffiliation = "member"
        self.is_user: bool = is_user
        self.is_system: bool = is_system

        self._nickname = nickname

        self.__update_jid(nickname)
        log.debug("Instantiation of: %r", self)

        self.contact: Optional["LegacyContact"] = None
        # we track if we already sent a presence for this participant.
        # if we didn't, we send it before the first message.
        # this way, event in plugins that don't map "user has joined" events,
        # we send a "join"-presence from the participant before the first message
        self.__presence_sent = False
        self.log = logging.getLogger(f"{self.user.bare_jid}:{self.jid}")

    def __repr__(self):
        return f"<Participant '{self.nickname}'/'{self.jid}' of '{self.muc}'>"

    @property
    def affiliation(self):
        return self._affiliation

    @affiliation.setter
    def affiliation(self, affiliation: MucAffiliation):
        if self._affiliation == affiliation:
            return
        self._affiliation = affiliation
        if not self.__presence_sent:
            return
        self.send_last_presence(force=True, no_cache_online=True)

    @property
    def role(self):
        return self._role

    @role.setter
    def role(self, role: MucRole):
        if self._role == role:
            return
        self._role = role
        if not self.__presence_sent:
            return
        self.send_last_presence(force=True, no_cache_online=True)

    def set_hats(self, hats: list[Hat]):
        if self._hats == hats:
            return
        self._hats = hats
        if not self.__presence_sent:
            return
        self.send_last_presence(force=True, no_cache_online=True)

    def __update_jid(self, unescaped_nickname: Optional[str]):
        j: JID = copy(self.muc.jid)

        if self.is_system:
            self.jid = j
            return

        nickname = unescaped_nickname

        if nickname:
            nickname = self._nickname_no_illegal = strip_illegal_chars(nickname)
        else:
            warnings.warn(
                "Only the system participant is allowed to not have a nickname"
            )
            nickname = f"unnamed-{uuid.uuid4()}"

        assert isinstance(nickname, str)

        try:
            # workaround for https://codeberg.org/poezio/slixmpp/issues/3480
            prohibit_output(nickname, [stringprep.in_table_a1])
            resourceprep(nickname)
        except (StringPrepError, StringprepError):
            nickname = nickname.encode("punycode").decode()

        # at this point there still might be control chars
        try:
            j.resource = nickname
        except InvalidJID:
            j.resource = strip_non_printable(nickname)

        if nickname != unescaped_nickname:
            self.muc._participants_by_escaped_nicknames[nickname] = self  # type:ignore

        self.jid = j

    def send_configuration_change(self, codes: tuple[int]):
        if not self.is_system:
            raise RuntimeError("This is only possible for the system participant")
        msg = self._make_message()
        msg["muc"]["status_codes"] = codes
        self._send(msg)

    @property
    def nickname(self):
        return self._nickname

    @nickname.setter
    def nickname(self, new_nickname: str):
        old = self._nickname
        if new_nickname == old:
            return

        cache = getattr(self, "_last_presence", None)
        if cache:
            last_seen = cache.last_seen
            kwargs = cache.presence_kwargs
        else:
            last_seen = None
            kwargs = {}

        kwargs["status_codes"] = {303}

        p = self._make_presence(ptype="unavailable", last_seen=last_seen, **kwargs)
        # in this order so pfrom=old resource and we actually use the escaped nick
        # in the muc/item/nick element
        self.__update_jid(new_nickname)
        p["muc"]["item"]["nick"] = self.jid.resource
        self._send(p)

        self._nickname = new_nickname

        kwargs["status_codes"] = set()
        p = self._make_presence(ptype="available", last_seen=last_seen, **kwargs)
        self.__add_nick_element(p)
        self._send(p)

        if old:
            self.muc.rename_participant(old, new_nickname)

    def _make_presence(
        self,
        *,
        last_seen: Optional[datetime] = None,
        status_codes: Optional[set[int]] = None,
        user_full_jid: Optional[JID] = None,
        **presence_kwargs,
    ):
        p = super()._make_presence(last_seen=last_seen, **presence_kwargs)
        p["muc"]["affiliation"] = self.affiliation
        p["muc"]["role"] = self.role
        if self._hats:
            p["hats"].add_hats(self._hats)
        codes = status_codes or set()
        if self.is_user:
            codes.add(110)
        if not self.muc.is_anonymous:
            if self.is_user:
                if user_full_jid:
                    p["muc"]["jid"] = user_full_jid
                else:
                    jid = copy(self.user.jid)
                    try:
                        jid.resource = next(
                            iter(self.muc.user_resources)  # type:ignore
                        )
                    except StopIteration:
                        jid.resource = "pseudo-resource"
                    p["muc"]["jid"] = self.user.jid
                codes.add(100)
            elif self.contact:
                p["muc"]["jid"] = self.contact.jid
                if a := self.contact.get_avatar():
                    p["vcard_temp_update"]["photo"] = a.id
            else:
                warnings.warn(
                    f"Private group but no 1:1 JID associated to '{self}'",
                )
        if self.is_user and (hash_ := self.session.avatar_hash):
            p["vcard_temp_update"]["photo"] = hash_
        p["muc"]["status_codes"] = codes
        return p

    @property
    def DISCO_NAME(self):
        return self.nickname

    def __send_presence_if_needed(
        self, stanza: Union[Message, Presence], full_jid: JID, archive_only: bool
    ):
        if (
            archive_only
            or self.is_system
            or self.is_user
            or self.__presence_sent
            or stanza["subject"]
        ):
            return
        if isinstance(stanza, Message):
            self.send_initial_presence(full_jid)

    @cached_property
    def __occupant_id(self):
        if self.contact:
            return self.contact.jid
        elif self.is_user:
            return "slidge-user"
        elif self.is_system:
            return "room"
        else:
            return str(uuid.uuid4())

    def _send(
        self,
        stanza: MessageOrPresenceTypeVar,
        full_jid: Optional[JID] = None,
        archive_only=False,
        **send_kwargs,
    ) -> MessageOrPresenceTypeVar:
        stanza["occupant-id"]["id"] = self.__occupant_id
        if isinstance(stanza, Presence):
            if stanza["type"] == "unavailable" and not self.__presence_sent:
                return stanza  # type:ignore
            self.__presence_sent = True
        if full_jid:
            stanza["to"] = full_jid
            self.__send_presence_if_needed(stanza, full_jid, archive_only)
            if self.is_user:
                assert stanza.stream is not None
                stanza.stream.send(stanza, use_filters=False)
            else:
                stanza.send()
        else:
            if isinstance(stanza, Message):
                self.muc.archive.add(stanza, self)
            if archive_only:
                return stanza
            for user_full_jid in self.muc.user_full_jids():
                stanza = copy(stanza)
                stanza["to"] = user_full_jid
                self.__send_presence_if_needed(stanza, user_full_jid, archive_only)
                stanza.send()
        return stanza

    def mucadmin_item(self):
        item = MUCAdminItem()
        item["nick"] = self.nickname
        item["affiliation"] = self.affiliation
        item["role"] = self.role
        if not self.muc.is_anonymous:
            if self.is_user:
                item["jid"] = self.user.bare_jid
            elif self.contact:
                item["jid"] = self.contact.jid.bare
            else:
                warnings.warn(
                    (
                        f"Public group but no contact JID associated to {self.jid} in"
                        f" {self}"
                    ),
                )
        return item

    def __add_nick_element(self, p: Presence):
        if (nick := self._nickname_no_illegal) != self.jid.resource:
            n = self.xmpp.plugin["xep_0172"].stanza.UserNick()
            n["nick"] = nick
            p.append(n)

    def _get_last_presence(self) -> Optional[CachedPresence]:
        own = super()._get_last_presence()
        if own is None and self.contact:
            return self.contact._get_last_presence()
        return own

    def send_initial_presence(
        self,
        full_jid: JID,
        nick_change=False,
        presence_id: Optional[str] = None,
    ):
        """
        Called when the user joins a MUC, as a mechanism
        to indicate to the joining XMPP client the list of "participants".

        Can be called this to trigger a "participant has joined the group" event.

        :param full_jid: Set this to only send to a specific user XMPP resource.
        :param nick_change: Used when the user joins and the MUC renames them (code 210)
        :param presence_id: set the presence ID. used internally by slidge
        """
        #  MUC status codes: https://xmpp.org/extensions/xep-0045.html#registrar-statuscodes
        codes = set()
        if nick_change:
            codes.add(210)

        if self.is_user:
            # the "initial presence" of the user has to be vanilla, as it is
            # a crucial part of the MUC join sequence for XMPP clients.
            kwargs = {}
        else:
            cache = self._get_last_presence()
            self.log.debug("Join muc, initial presence: %s", cache)
            if cache:
                ptype = cache.ptype
                if ptype == "unavailable":
                    return
                kwargs = dict(
                    last_seen=cache.last_seen, pstatus=cache.pstatus, pshow=cache.pshow
                )
            else:
                kwargs = {}
        p = self._make_presence(
            status_codes=codes,
            user_full_jid=full_jid,
            **kwargs,  # type:ignore
        )
        if presence_id:
            p["id"] = presence_id
        self.__add_nick_element(p)
        self._send(p, full_jid)

    def leave(self):
        """
        Call this when the participant leaves the room
        """
        self.muc.remove_participant(self)

    def kick(self):
        """
        Call this when the participant is kicked from the room
        """
        self.muc.remove_participant(self, kick=True)

    def ban(self):
        """
        Call this when the participant is banned from the room
        """
        self.muc.remove_participant(self, ban=True)

    def get_disco_info(self, jid: OptJid = None, node: Optional[str] = None):
        if self.contact is not None:
            return self.contact.get_disco_info()
        return super().get_disco_info()

    def moderate(self, legacy_msg_id: LegacyMessageType, reason: Optional[str] = None):
        m = self.muc.get_system_participant()._make_message()
        m["apply_to"]["id"] = self._legacy_to_xmpp(legacy_msg_id)
        m["apply_to"]["moderated"].enable("retract")
        m["apply_to"]["moderated"]["by"] = self.jid
        if reason:
            m["apply_to"]["moderated"]["reason"] = reason
        self._send(m)

    def set_room_subject(
        self,
        subject: str,
        full_jid: Optional[JID] = None,
        when: Optional[datetime] = None,
        update_muc=True,
    ):
        if update_muc:
            self.muc._subject = subject  # type: ignore
            self.muc.subject_setter = self
            self.muc.subject_date = when

        msg = self._make_message()
        if when is not None:
            msg["delay"].set_stamp(when)
            msg["delay"]["from"] = self.muc.jid
        msg["subject"] = subject or str(self.muc.name)
        self._send(msg, full_jid)


log = logging.getLogger(__name__)



import logging
import typing
from collections import defaultdict

from slixmpp import JID, Iq, Message
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0356.permissions import (
    MessagePermission,
    Permissions,
    RosterAccess,
)
from slixmpp.types import JidStr
from slixmpp.xmlstream import StanzaBase
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath

from . import stanza

log = logging.getLogger(__name__)


# noinspection PyPep8Naming
class XEP_0356_OLD(BasePlugin):
    """
    XEP-0356: Privileged Entity

    Events:

    ::

        privileges_advertised_old -- Received message/privilege from the server
    """

    name = "xep_0356_old"
    description = "XEP-0356: Privileged Entity (slidge - old namespace)"
    dependencies = {"xep_0297"}
    stanza = stanza

    granted_privileges: defaultdict[JidStr, Permissions] = defaultdict(Permissions)

    def plugin_init(self):
        if not self.xmpp.is_component:
            log.error("XEP 0356 is only available for components")
            return

        stanza.register()

        self.xmpp.register_handler(
            Callback(
                "Privileges_old",
                StanzaPath("message/privilege_old"),
                self._handle_privilege,
            )
        )

    def plugin_end(self):
        self.xmpp.remove_handler("Privileges_old")

    def _handle_privilege(self, msg: StanzaBase):
        """
        Called when the XMPP server advertise the component's privileges.

        Stores the privileges in this instance's granted_privileges attribute (a dict)
        and raises the privileges_advertised event
        """
        for perm in msg["privilege_old"]["perms"]:
            setattr(
                self.granted_privileges[msg.get_from()], perm["access"], perm["type"]
            )
        log.debug(f"Privileges (old): {self.granted_privileges}")
        self.xmpp.event("privileges_advertised_old")

    def send_privileged_message(self, msg: Message):
        if (
            self.granted_privileges[msg.get_from().domain].message
            != MessagePermission.OUTGOING
        ):
            raise PermissionError(
                "The server hasn't authorized us to send messages on behalf of other users"
            )
        else:
            self._make_privileged_message(msg).send()

    def _make_privileged_message(self, msg: Message):
        server = msg.get_from().domain
        wrapped = self.xmpp.make_message(mto=server, mfrom=self.xmpp.boundjid.bare)
        wrapped["privilege_old"]["forwarded"].append(msg)
        return wrapped

    def _make_get_roster(self, jid: typing.Union[JID, str], **iq_kwargs):
        return self.xmpp.make_iq_get(
            queryxmlns="jabber:iq:roster",
            ifrom=self.xmpp.boundjid.bare,
            ito=jid,
            **iq_kwargs,
        )

    def _make_set_roster(
        self,
        jid: typing.Union[JID, str],
        roster_items: dict,
        **iq_kwargs,
    ):
        iq = self.xmpp.make_iq_set(
            ifrom=self.xmpp.boundjid.bare,
            ito=jid,
            **iq_kwargs,
        )
        iq["roster"]["items"] = roster_items
        return iq

    async def get_roster(self, jid: typing.Union[JID, str], **send_kwargs) -> Iq:
        """
        Return the roster of user on the server the component has privileged access to.

        Raises ValueError if the server did not advertise the corresponding privileges

        :param jid: user we want to fetch the roster from
        """
        if isinstance(jid, str):
            jid = JID(jid)
        if self.granted_privileges[jid.domain].roster not in (
            RosterAccess.GET,
            RosterAccess.BOTH,
        ):
            raise PermissionError(
                "The server did not grant us privileges to get rosters"
            )
        else:
            return await self._make_get_roster(jid).send(**send_kwargs)

    async def set_roster(
        self, jid: typing.Union[JID, str], roster_items: dict, **send_kwargs
    ) -> Iq:
        """
        Return the roster of user on the server the component has privileged access to.

        Raises ValueError if the server did not advertise the corresponding privileges

        :param jid: user we want to add or modify roster items
        :param roster_items: a dict containing the roster items' JIDs as keys and
            nested dicts containing names, subscriptions and groups.
            Example:
            {
                "friend1@example.com": {
                    "name": "Friend 1",
                    "subscription": "both",
                    "groups": ["group1", "group2"],
                },
                "friend2@example.com": {
                    "name": "Friend 2",
                    "subscription": "from",
                    "groups": ["group3"],
            },
        }
        """
        if isinstance(jid, str):
            jid = JID(jid)
        if self.granted_privileges[jid.domain].roster not in (
            RosterAccess.GET,
            RosterAccess.BOTH,
        ):
            raise PermissionError(
                "The server did not grant us privileges to set rosters"
            )
        else:
            return await self._make_set_roster(jid, roster_items).send(**send_kwargs)



# This module contains patches for slixmpp; some have pending requests upstream
# and should be removed on the next slixmpp release.

# ruff: noqa: F401

import slixmpp.plugins
from slixmpp import Message
from slixmpp.plugins.xep_0050 import XEP_0050, Command
from slixmpp.xmlstream import StanzaBase

from . import (  # xep_0356,
    link_preview,
    xep_0077,
    xep_0100,
    xep_0153,
    xep_0264,
    xep_0292,
    xep_0313,
    xep_0317,
    xep_0356_old,
    xep_0424,
    xep_0490,
)


def session_bind(self, jid):
    self.xmpp["xep_0030"].add_feature(Command.namespace)
    # awful hack to for the disco items: we need to comment this line
    # related issue: https://todo.sr.ht/~nicoco/slidge/131
    # self.xmpp['xep_0030'].set_items(node=Command.namespace, items=tuple())


XEP_0050.session_bind = session_bind  # type:ignore


def reply(self, body=None, clear=True):
    """
    Overrides slixmpp's Message.reply(), since it strips to sender's resource
    for mtype=groupchat, and we do not want that, because when we raise an XMPPError,
    we actually want to preserve the resource.
    (this is called in RootStanza.exception() to handle XMPPErrors)
    """
    new_message = StanzaBase.reply(self, clear)
    new_message["thread"] = self["thread"]
    new_message["parent_thread"] = self["parent_thread"]

    del new_message["id"]
    if self.stream is not None and self.stream.use_message_ids:
        new_message["id"] = self.stream.new_id()

    if body is not None:
        new_message["body"] = body
    return new_message


slixmpp.plugins.PLUGINS.extend(
    [
        "link_preview",
        "xep_0264",
        "xep_0292_provider",
        "xep_0317",
        "xep_0356_old",
        "xep_0490",
    ]
)


Message.reply = reply  # type: ignore



import functools
import logging
import os
import re
import shutil
import stat
import tempfile
import warnings
from datetime import datetime
from mimetypes import guess_type
from pathlib import Path
from typing import IO, Collection, Optional, Sequence, Union
from urllib.parse import quote as urlquote
from uuid import uuid4
from xml.etree import ElementTree as ET

import blurhash
from PIL import Image
from slixmpp import JID, Message
from slixmpp.exceptions import IqError
from slixmpp.plugins.xep_0363 import FileUploadError
from slixmpp.plugins.xep_0385.stanza import Sims
from slixmpp.plugins.xep_0447.stanza import StatelessFileSharing

from ...util.sql import db
from ...util.types import (
    LegacyAttachment,
    LegacyMessageType,
    LegacyThreadType,
    MessageReference,
)
from ...util.util import fix_suffix
from .. import config
from ..cache import avatar_cache
from .message_maker import MessageMaker


class AttachmentMixin(MessageMaker):
    def send_text(self, *_, **k) -> Optional[Message]:
        raise NotImplementedError

    async def __upload(
        self,
        file_path: Path,
        file_name: Optional[str] = None,
        content_type: Optional[str] = None,
    ):
        if file_name and file_path.name != file_name:
            d = Path(tempfile.mkdtemp())
            temp = d / file_name
            temp.symlink_to(file_path)
            file_path = temp
        else:
            d = None
        if config.UPLOAD_SERVICE:
            domain = None
        else:
            domain = re.sub(r"^.*?\.", "", self.xmpp.boundjid.bare)
        try:
            new_url = await self.xmpp.plugin["xep_0363"].upload_file(
                filename=file_path,
                content_type=content_type,
                ifrom=config.UPLOAD_REQUESTER or self.xmpp.boundjid,
                domain=JID(domain),
            )
        except (FileUploadError, IqError) as e:
            warnings.warn(f"Something is wrong with the upload service: {e!r}")
            return None
        finally:
            if d is not None:
                file_path.unlink()
                d.rmdir()

        return new_url

    @staticmethod
    async def __no_upload(
        file_path: Path,
        file_name: Optional[str] = None,
        legacy_file_id: Optional[Union[str, int]] = None,
    ):
        file_id = str(uuid4()) if legacy_file_id is None else str(legacy_file_id)
        assert config.NO_UPLOAD_PATH is not None
        assert config.NO_UPLOAD_URL_PREFIX is not None
        destination_dir = Path(config.NO_UPLOAD_PATH) / file_id

        if destination_dir.exists():
            log.debug("Dest dir exists: %s", destination_dir)
            files = list(f for f in destination_dir.glob("**/*") if f.is_file())
            if len(files) == 1:
                log.debug(
                    "Found the legacy attachment '%s' at '%s'",
                    legacy_file_id,
                    files[0],
                )
                name = files[0].name
                uu = files[0].parent.name  # anti-obvious url trick, see below
                return files[0], "/".join([file_id, uu, name])
            else:
                log.warning(
                    (
                        "There are several or zero files in %s, "
                        "slidge doesn't know which one to pick among %s. "
                        "Removing the dir."
                    ),
                    destination_dir,
                    files,
                )
                shutil.rmtree(destination_dir)

        log.debug("Did not find a file in: %s", destination_dir)
        # let's use a UUID to avoid URLs being too obvious
        uu = str(uuid4())
        destination_dir = destination_dir / uu
        destination_dir.mkdir(parents=True)

        name = file_name or file_path.name
        destination = destination_dir / name
        method = config.NO_UPLOAD_METHOD
        if method == "copy":
            shutil.copy2(file_path, destination)
        elif method == "hardlink":
            os.link(file_path, destination)
        elif method == "symlink":
            os.symlink(file_path, destination, target_is_directory=True)
        elif method == "move":
            shutil.move(file_path, destination)
        else:
            raise RuntimeError("No upload method not recognized", method)

        if config.NO_UPLOAD_FILE_READ_OTHERS:
            log.debug("Changing perms of %s", destination)
            destination.chmod(destination.stat().st_mode | stat.S_IROTH)
        uploaded_url = "/".join([file_id, uu, name])

        return destination, uploaded_url

    async def __get_url(
        self,
        file_path: Optional[Path] = None,
        data_stream: Optional[IO[bytes]] = None,
        data: Optional[bytes] = None,
        file_url: Optional[str] = None,
        file_name: Optional[str] = None,
        content_type: Optional[str] = None,
        legacy_file_id: Optional[Union[str, int]] = None,
    ) -> tuple[bool, Optional[Path], str]:
        if legacy_file_id:
            cache = db.attachment_get_url(legacy_file_id)
            if cache is not None:
                async with self.session.http.head(cache) as r:
                    if r.status < 400:
                        return False, None, cache
                    else:
                        db.attachment_remove(legacy_file_id)

        if file_url and config.USE_ATTACHMENT_ORIGINAL_URLS:
            return False, None, file_url

        if file_name and len(file_name) > config.ATTACHMENT_MAXIMUM_FILE_NAME_LENGTH:
            log.debug("Trimming long filename: %s", file_name)
            base, ext = os.path.splitext(file_name)
            file_name = (
                base[: config.ATTACHMENT_MAXIMUM_FILE_NAME_LENGTH - len(ext)] + ext
            )

        if file_path is None:
            file_name = str(uuid4()) if file_name is None else file_name
            temp_dir = Path(tempfile.mkdtemp())
            file_path = temp_dir / file_name
            if file_url:
                async with self.session.http.get(file_url) as r:
                    with file_path.open("wb") as f:
                        f.write(await r.read())

            else:
                if data_stream is not None:
                    data = data_stream.read()
                if data is None:
                    raise RuntimeError

                with file_path.open("wb") as f:
                    f.write(data)

            is_temp = not bool(config.NO_UPLOAD_PATH)
        else:
            is_temp = False

        if config.FIX_FILENAME_SUFFIX_MIME_TYPE:
            file_name = str(fix_suffix(file_path, content_type, file_name))

        if config.NO_UPLOAD_PATH:
            local_path, new_url = await self.__no_upload(
                file_path, file_name, legacy_file_id
            )
            new_url = (config.NO_UPLOAD_URL_PREFIX or "") + "/" + urlquote(new_url)
        else:
            local_path = file_path
            new_url = await self.__upload(file_path, file_name, content_type)
        if legacy_file_id:
            db.attachment_store_url(legacy_file_id, new_url)

        return is_temp, local_path, new_url

    async def __set_sims(
        self,
        msg: Message,
        uploaded_url: str,
        path: Optional[Path],
        content_type: Optional[str] = None,
        caption: Optional[str] = None,
        file_name: Optional[str] = None,
    ):
        cache = db.attachment_get_sims(uploaded_url)
        if cache:
            msg.append(Sims(xml=ET.fromstring(cache)))
            return

        if not path:
            return

        sims = self.xmpp["xep_0385"].get_sims(
            path, [uploaded_url], content_type, caption
        )
        if file_name:
            sims["sims"]["file"]["name"] = file_name
        if content_type is not None and content_type.startswith("image"):
            try:
                h, x, y = await self.xmpp.loop.run_in_executor(
                    avatar_cache._thread_pool, get_blurhash, path
                )
            except Exception as e:
                log.debug("Could not generate a blurhash", exc_info=e)
            else:
                thumbnail = sims["sims"]["file"]["thumbnail"]
                thumbnail["width"] = x
                thumbnail["height"] = y
                thumbnail["media-type"] = "image/blurhash"
                thumbnail["uri"] = "data:image/blurhash," + urlquote(h)

        db.attachment_store_sims(uploaded_url, str(sims))

        msg.append(sims)

    def __set_sfs(
        self,
        msg: Message,
        uploaded_url: str,
        path: Optional[Path],
        content_type: Optional[str] = None,
        caption: Optional[str] = None,
        file_name: Optional[str] = None,
    ):
        cache = db.attachment_get_sfs(uploaded_url)
        if cache:
            msg.append(StatelessFileSharing(xml=ET.fromstring(cache)))
            return

        if not path:
            return

        sfs = self.xmpp["xep_0447"].get_sfs(path, [uploaded_url], content_type, caption)
        if file_name:
            sfs["file"]["name"] = file_name
        db.attachment_store_sfs(uploaded_url, str(sfs))

        msg.append(sfs)

    def __send_url(
        self,
        msg: Message,
        legacy_msg_id: LegacyMessageType,
        uploaded_url: str,
        caption: Optional[str] = None,
        carbon=False,
        when: Optional[datetime] = None,
        **kwargs,
    ) -> list[Message]:
        msg["oob"]["url"] = uploaded_url
        msg["body"] = uploaded_url
        if caption:
            m1 = self._send(msg, carbon=carbon, **kwargs)
            m2 = self.send_text(
                caption, legacy_msg_id=legacy_msg_id, when=when, carbon=carbon, **kwargs
            )
            return [m1, m2] if m2 else [m1]
        else:
            self._set_msg_id(msg, legacy_msg_id)
            return [self._send(msg, carbon=carbon, **kwargs)]

    async def send_file(
        self,
        file_path: Optional[Union[Path, str]] = None,
        legacy_msg_id: Optional[LegacyMessageType] = None,
        *,
        data_stream: Optional[IO[bytes]] = None,
        data: Optional[bytes] = None,
        file_url: Optional[str] = None,
        file_name: Optional[str] = None,
        content_type: Optional[str] = None,
        reply_to: Optional[MessageReference] = None,
        when: Optional[datetime] = None,
        caption: Optional[str] = None,
        legacy_file_id: Optional[Union[str, int]] = None,
        thread: Optional[LegacyThreadType] = None,
        **kwargs,
    ) -> tuple[Optional[str], list[Message]]:
        """
        Send a single file from this :term:`XMPP Entity`.

        :param file_path: Path to the attachment
        :param data_stream: Alternatively, a stream of bytes (such as a File object)
        :param data: Alternatively, a bytes object
        :param file_url: Alternatively, a URL
        :param file_name: How the file should be named.
        :param content_type: MIME type, inferred from filename if not given
        :param legacy_msg_id: If you want to be able to transport read markers from the gateway
            user to the legacy network, specify this
        :param reply_to: Quote another message (:xep:`0461`)
        :param when: when the file was sent, for a "delay" tag (:xep:`0203`)
        :param caption: an optional text that is linked to the file
        :param legacy_file_id: A unique identifier for the file on the legacy network.
             Plugins should try their best to provide it, to avoid duplicates.
        :param thread:
        """
        carbon = kwargs.pop("carbon", False)
        mto = kwargs.pop("mto", None)
        store_multi = kwargs.pop("store_multi", True)
        msg = self._make_message(
            when=when,
            reply_to=reply_to,
            carbon=carbon,
            mto=mto,
            thread=thread,
        )

        if content_type is None and (name := (file_name or file_path or file_url)):
            content_type, _ = guess_type(name)

        is_temp, local_path, new_url = await self.__get_url(
            Path(file_path) if file_path else None,
            data_stream,
            data,
            file_url,
            file_name,
            content_type,
            legacy_file_id,
        )

        if new_url is None:
            msg["body"] = (
                "I tried to send a file, but something went wrong. "
                "Tell your slidge admin to check the logs."
            )
            self._set_msg_id(msg, legacy_msg_id)
            return None, [self._send(msg, **kwargs)]

        await self.__set_sims(
            msg, new_url, local_path, content_type, caption, file_name
        )
        self.__set_sfs(msg, new_url, local_path, content_type, caption, file_name)
        if is_temp and isinstance(local_path, Path):
            local_path.unlink()
            local_path.parent.rmdir()

        msgs = self.__send_url(
            msg, legacy_msg_id, new_url, caption, carbon, when, **kwargs
        )
        if store_multi:
            self.__store_multi(legacy_msg_id, msgs)
        return new_url, msgs

    def __send_body(
        self,
        body: Optional[str] = None,
        legacy_msg_id: Optional[LegacyMessageType] = None,
        reply_to: Optional[MessageReference] = None,
        when: Optional[datetime] = None,
        thread: Optional[LegacyThreadType] = None,
        **kwargs,
    ) -> Optional[Message]:
        if body:
            return self.send_text(
                body,
                legacy_msg_id,
                reply_to=reply_to,
                when=when,
                thread=thread,
                **kwargs,
            )
        else:
            return None

    async def send_files(
        self,
        attachments: Collection[LegacyAttachment],
        legacy_msg_id: Optional[LegacyMessageType] = None,
        body: Optional[str] = None,
        *,
        reply_to: Optional[MessageReference] = None,
        when: Optional[datetime] = None,
        thread: Optional[LegacyThreadType] = None,
        body_first=False,
        correction=False,
        correction_event_id: Optional[LegacyMessageType] = None,
        **kwargs,
    ):
        # TODO: once the epic XEP-0385 vs XEP-0447 battle is over, pick
        #       one and stop sending several attachments this way
        # we attach the legacy_message ID to the last message we send, because
        # we don't want several messages with the same ID (especially for MUC MAM)
        # TODO: refactor this so we limit the number of SQL calls, ie, if
        #       the legacy file ID is known, only fetch the row once, and if it
        #       is new, write it all in a single call
        if not attachments and not body:
            # ignoring empty message
            return
        send_body = functools.partial(
            self.__send_body,
            body=body,
            reply_to=reply_to,
            when=when,
            thread=thread,
            correction=correction,
            legacy_msg_id=legacy_msg_id,
            correction_event_id=correction_event_id,
            **kwargs,
        )
        all_msgs = []
        if body_first:
            all_msgs.append(send_body())
        last_attachment_i = len(attachments) - 1
        for i, attachment in enumerate(attachments):
            last = i == last_attachment_i
            if last and not body:
                legacy = legacy_msg_id
            else:
                legacy = None
            _url, msgs = await self.send_file(
                file_path=attachment.path,
                legacy_msg_id=legacy,
                file_url=attachment.url,
                data_stream=attachment.stream,
                data=attachment.data,
                reply_to=reply_to,
                when=when,
                thread=thread,
                file_name=attachment.name,
                content_type=attachment.content_type,
                legacy_file_id=attachment.legacy_file_id,
                caption=attachment.caption,
                store_multi=False,
                **kwargs,
            )
            all_msgs.extend(msgs)
        if not body_first:
            all_msgs.append(send_body())
        self.__store_multi(legacy_msg_id, all_msgs)

    def __store_multi(
        self,
        legacy_msg_id: Optional[LegacyMessageType],
        all_msgs: Sequence[Optional[Message]],
    ):
        if legacy_msg_id is None:
            return
        ids = []
        for msg in all_msgs:
            if not msg:
                continue
            if stanza_id := msg.get_plugin("stanza_id", check=True):
                ids.append(stanza_id["id"])
            else:
                ids.append(msg.get_id())
        db.attachment_store_legacy_to_multi_xmpp_msg_ids(legacy_msg_id, ids)


def get_blurhash(path: Path, n=9) -> tuple[str, int, int]:
    img = Image.open(path)
    width, height = img.size
    n = min(width, height, n)
    if width == height:
        x = y = n
    elif width > height:
        x = n
        y = round(n * height / width)
    else:
        x = round(n * width / height)
        y = n
    # There are 2 blurhash-python packages:
    # https://github.com/woltapp/blurhash-python
    # https://github.com/halcy/blurhash-python
    # With this hack we're compatible with both, which is useful for packaging
    # without using pyproject.toml, as most distro do
    try:
        hash_ = blurhash.encode(img, x, y)
    except TypeError:
        # We are using halcy's blurhash which expects
        # the 1st argument to be a 3-dimensional array
        import numpy  # type:ignore

        hash_ = blurhash.encode(numpy.array(img.convert("RGB")), x, y)
    return hash_, width, height


log = logging.getLogger(__name__)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission
import logging
from asyncio import Future
from collections.abc import AsyncGenerator
from datetime import datetime
from typing import Any, Awaitable, Callable, Dict, Optional, Tuple

from slixmpp import JID
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0004.stanza import Form
from slixmpp.stanza import Iq, Message
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Collector
from slixmpp.xmlstream.matcher import MatchXMLMask

from . import stanza

log = logging.getLogger(__name__)


class XEP_0313(BasePlugin):
    """
    XEP-0313 Message Archive Management
    """

    name = "xep_0313"
    description = "XEP-0313: Message Archive Management"
    dependencies = {"xep_0004", "xep_0030", "xep_0050", "xep_0059", "xep_0297"}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(stanza.MAM, Form)
        register_stanza_plugin(Iq, stanza.MAM)
        register_stanza_plugin(Message, stanza.Result)
        register_stanza_plugin(Iq, stanza.Fin)
        register_stanza_plugin(stanza.Result, self.xmpp["xep_0297"].stanza.Forwarded)
        register_stanza_plugin(stanza.MAM, self.xmpp["xep_0059"].stanza.Set)
        register_stanza_plugin(stanza.Fin, self.xmpp["xep_0059"].stanza.Set)
        register_stanza_plugin(Iq, stanza.Metadata)
        register_stanza_plugin(stanza.Metadata, stanza.Start)
        register_stanza_plugin(stanza.Metadata, stanza.End)

    def retrieve(
        self,
        jid: Optional[JID] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        with_jid: Optional[JID] = None,
        ifrom: Optional[JID] = None,
        reverse: bool = False,
        timeout: int = None,
        callback: Callable[[Iq], None] = None,
        iterator: bool = False,
        rsm: Optional[Dict[str, Any]] = None,
    ) -> Awaitable:
        """
        Send a MAM query and retrieve the results.

        :param JID jid: Entity holding the MAM records
        :param datetime start,end: MAM query temporal boundaries
        :param JID with_jid: Filter results on this JID
        :param JID ifrom: To change the from address of the query
        :param bool reverse: Get the results in reverse order
        :param int timeout: IQ timeout
        :param func callback: Custom callback for handling results
        :param bool iterator: Use RSM and iterate over a paginated query
        :param dict rsm: RSM custom options
        """
        iq, stanza_mask = self._pre_mam_retrieve(jid, start, end, with_jid, ifrom)
        query_id = iq["id"]
        amount = 10
        if rsm:
            for key, value in rsm.items():
                iq["mam"]["rsm"][key] = str(value)
                if key == "max":
                    amount = value
        cb_data = {}

        xml_mask = str(stanza_mask)

        def pre_cb(query: Iq) -> None:
            stanza_mask["mam_result"]["queryid"] = query["id"]
            xml_mask = str(stanza_mask)
            query["mam"]["queryid"] = query["id"]
            collector = Collector("MAM_Results_%s" % query_id, MatchXMLMask(xml_mask))
            self.xmpp.register_handler(collector)
            cb_data["collector"] = collector

        def post_cb(result: Iq) -> None:
            results = cb_data["collector"].stop()
            if result["type"] == "result":
                result["mam"]["results"] = results
                result["mam_fin"]["results"] = results

        if iterator:
            return self.xmpp["xep_0059"].iterate(
                iq,
                "mam",
                "results",
                amount=amount,
                reverse=reverse,
                recv_interface="mam_fin",
                pre_cb=pre_cb,
                post_cb=post_cb,
            )

        collector = Collector("MAM_Results_%s" % query_id, MatchXMLMask(xml_mask))
        self.xmpp.register_handler(collector)

        def wrapped_cb(iq: Iq) -> None:
            results = collector.stop()
            if iq["type"] == "result":
                iq["mam"]["results"] = results
            if callback:
                callback(iq)

        return iq.send(timeout=timeout, callback=wrapped_cb)

    async def iterate(
        self,
        jid: Optional[JID] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        with_jid: Optional[JID] = None,
        ifrom: Optional[JID] = None,
        reverse: bool = False,
        rsm: Optional[Dict[str, Any]] = None,
        total: Optional[int] = None,
    ) -> AsyncGenerator:
        """
        Iterate over each message of MAM query.

        .. versionadded:: 1.8.0

        :param jid: Entity holding the MAM records
        :param start: MAM query start time
        :param end: MAM query end time
        :param with_jid: Filter results on this JID
        :param ifrom: To change the from address of the query
        :param reverse: Get the results in reverse order
        :param rsm: RSM custom options
        :param total: A number of messages received after which the query
                      should stop.
        """
        iq, stanza_mask = self._pre_mam_retrieve(jid, start, end, with_jid, ifrom)
        query_id = iq["id"]
        amount = 10

        if rsm:
            for key, value in rsm.items():
                iq["mam"]["rsm"][key] = str(value)
                if key == "max":
                    amount = value
        cb_data = {}

        def pre_cb(query: Iq) -> None:
            stanza_mask["mam_result"]["queryid"] = query["id"]
            xml_mask = str(stanza_mask)
            query["mam"]["queryid"] = query["id"]
            collector = Collector("MAM_Results_%s" % query_id, MatchXMLMask(xml_mask))
            self.xmpp.register_handler(collector)
            cb_data["collector"] = collector

        def post_cb(result: Iq) -> None:
            results = cb_data["collector"].stop()
            if result["type"] == "result":
                result["mam"]["results"] = results
                result["mam_fin"]["results"] = results

        iterator = self.xmpp["xep_0059"].iterate(
            iq,
            "mam",
            "results",
            amount=amount,
            reverse=reverse,
            recv_interface="mam_fin",
            pre_cb=pre_cb,
            post_cb=post_cb,
        )
        recv_count = 0

        async for page in iterator:
            messages = [message for message in page["mam"]["results"]]
            if reverse:
                messages.reverse()
            for message in messages:
                yield message
                recv_count += 1
                if total is not None and recv_count >= total:
                    break
            if total is not None and recv_count >= total:
                break

    def _pre_mam_retrieve(
        self,
        jid: Optional[JID] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        with_jid: Optional[JID] = None,
        ifrom: Optional[JID] = None,
    ) -> Tuple[Iq, Message]:
        """Build the IQ and stanza mask for MAM results"""
        iq = self.xmpp.make_iq_set(ito=jid, ifrom=ifrom)
        query_id = iq["id"]
        iq["mam"]["queryid"] = query_id
        iq["mam"]["start"] = start
        iq["mam"]["end"] = end
        iq["mam"]["with"] = with_jid

        stanza_mask = self.xmpp.Message()

        auto_origin = stanza_mask.xml.find("{urn:xmpp:sid:0}origin-id")
        if auto_origin is not None:
            stanza_mask.xml.remove(auto_origin)
        del stanza_mask["id"]
        del stanza_mask["lang"]
        stanza_mask["from"] = jid
        stanza_mask["mam_result"]["queryid"] = query_id

        return (iq, stanza_mask)

    async def get_fields(self, jid: Optional[JID] = None, **iqkwargs) -> Form:
        """Get MAM query fields.

        .. versionadded:: 1.8.0

        :param jid: JID to retrieve the policy from.
        :return: The Form of allowed options
        """
        ifrom = iqkwargs.pop("ifrom", None)
        iq = self.xmpp.make_iq_get(ito=jid, ifrom=ifrom)
        iq.enable("mam")
        result = await iq.send(**iqkwargs)
        return result["mam"]["form"]

    async def get_configuration_commands(
        self, jid: Optional[JID], **discokwargs
    ) -> Future:
        """Get the list of MAM advanced configuration commands.

        .. versionchanged:: 1.8.0

        :param jid: JID to get the commands from.
        """
        if jid is None:
            jid = self.xmpp.boundjid.bare
        return await self.xmpp["xep_0030"].get_items(
            jid=jid, node="urn:xmpp:mam#configure", **discokwargs
        )

    def get_archive_metadata(self, jid: Optional[JID] = None, **iqkwargs) -> Future:
        """Get the archive metadata from a JID.

        :param jid: JID to get the metadata from.
        """
        ifrom = iqkwargs.pop("ifrom", None)
        iq = self.xmpp.make_iq_get(ito=jid, ifrom=ifrom)
        iq.enable("mam_metadata")
        return iq.send(**iqkwargs)



from abc import ABC
from dataclasses import dataclass, field
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Collection,
    Iterable,
    Optional,
    Type,
    TypedDict,
    Union,
)

from slixmpp import JID  # type: ignore[attr-defined]
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0004 import Form as SlixForm  # type: ignore[attr-defined]
from slixmpp.plugins.xep_0004 import (
    FormField as SlixFormField,  # type: ignore[attr-defined]
)
from slixmpp.types import JidStr

from ..core import config
from ..util.db import user_store
from ..util.types import AnyBaseSession, FieldType

if TYPE_CHECKING:
    from ..core.gateway import BaseGateway
    from ..core.session import BaseSession


HandlerType = Union[
    Callable[[AnyBaseSession, JID], "CommandResponseType"],
    Callable[[AnyBaseSession, JID], Awaitable["CommandResponseType"]],
]

FormValues = dict[str, Union[str, JID, bool]]


FormHandlerType = Callable[
    [FormValues, AnyBaseSession, JID],
    Awaitable["CommandResponseType"],
]

ConfirmationHandlerType = Callable[
    [Optional[AnyBaseSession], JID], Awaitable["CommandResponseType"]
]


@dataclass
class TableResult:
    """
    Structured data as the result of a command
    """

    fields: Collection["FormField"]
    """
    The 'columns names' of the table.
    """
    items: Collection[dict[str, Union[str, JID]]]
    """
    The rows of the table. Each row is a dict where keys are the fields ``var``
    attribute.
    """
    description: str
    """
    A description of the content of the table.
    """

    jids_are_mucs: bool = False

    def get_xml(self) -> SlixForm:
        """
        Get a slixmpp "form" (with <reported> header)to represent the data

        :return: some XML
        """
        form = SlixForm()  # type: ignore[no-untyped-call]
        form["type"] = "result"
        form["title"] = self.description
        for f in self.fields:
            form.add_reported(f.var, label=f.label, type=f.type)  # type: ignore[no-untyped-call]
        for item in self.items:
            form.add_item({k: str(v) for k, v in item.items()})  # type: ignore[no-untyped-call]
        return form


@dataclass
class SearchResult(TableResult):
    """
    Results of the search command (search for contacts via Jabber Search)

    Return type of :meth:`BaseSession.search`.
    """

    description: str = "Contact search results"


@dataclass
class Confirmation:
    """
    A confirmation 'dialog'
    """

    prompt: str
    """
    The text presented to the command triggering user
    """
    handler: ConfirmationHandlerType
    """
    An async function that should return a ResponseType
    """
    success: Optional[str] = None
    """
    Text in case of success, used if handler does not return anything
    """
    handler_args: Iterable[Any] = field(default_factory=list)
    """
    arguments passed to the handler
    """
    handler_kwargs: dict[str, Any] = field(default_factory=dict)
    """
    keyword arguments passed to the handler
    """

    def get_form(self) -> SlixForm:
        """
        Get the slixmpp form

        :return: some xml
        """
        form = SlixForm()  # type: ignore[no-untyped-call]
        form["type"] = "form"
        form["title"] = self.prompt
        form.append(
            FormField(
                "confirm", type="boolean", value="true", label="Confirm"
            ).get_xml()
        )
        return form


@dataclass
class Form:
    """
    A form, to request user input
    """

    title: str
    instructions: str
    fields: Collection["FormField"]
    handler: FormHandlerType
    handler_args: Iterable[Any] = field(default_factory=list)
    handler_kwargs: dict[str, Any] = field(default_factory=dict)

    def get_values(
        self, slix_form: SlixForm
    ) -> dict[str, Union[list[str], list[JID], str, JID, bool, None]]:
        """
        Parse form submission

        :param slix_form: the xml received as the submission of a form
        :return: A dict where keys=field.var and values are either strings
            or JIDs (if field.type=jid-single)
        """
        str_values: dict[str, str] = slix_form.get_values()  # type: ignore[no-untyped-call]
        values = {}
        for f in self.fields:
            values[f.var] = f.validate(str_values.get(f.var))
        return values

    def get_xml(self) -> SlixForm:
        """
        Get the slixmpp "form"

        :return: some XML
        """
        form = SlixForm()  # type: ignore[no-untyped-call]
        form["type"] = "form"
        form["instructions"] = self.instructions
        form["title"] = self.title
        for fi in self.fields:
            form.append(fi.get_xml())
        return form


class CommandAccess(int, Enum):
    """
    Defines who can access a given Command
    """

    ADMIN_ONLY = 0
    USER = 1
    USER_LOGGED = 2
    USER_NON_LOGGED = 3
    NON_USER = 4
    ANY = 5


class Option(TypedDict):
    """
    Options to be used for ``FormField``s of type ``list-*``
    """

    label: str
    value: str


# TODO: support forms validation XEP-0122
@dataclass
class FormField:
    """
    Represents a field of the form that a user will see when registering to the gateway
    via their XMPP client.
    """

    var: str = ""
    """
    Internal name of the field, will be used to retrieve via :py:attr:`slidge.GatewayUser.registration_form`
    """
    label: Optional[str] = None
    """Description of the field that the user will see"""
    required: bool = False
    """Whether this field is mandatory or not"""
    private: bool = False
    """
    For sensitive info that should not be displayed on screen while the user types.
    Forces field_type to "text-private"
    """
    type: FieldType = "text-single"
    """Type of the field, see `XEP-0004 <https://xmpp.org/extensions/xep-0004.html#protocol-fieldtypes>`_"""
    value: str = ""
    """Pre-filled value. Will be automatically pre-filled if a registered user modifies their subscription"""
    options: Optional[list[Option]] = None

    image_url: Optional[str] = None
    """An image associated to this field, eg, a QR code"""

    def __post_init__(self) -> None:
        if self.private:
            self.type = "text-private"

    def __acceptable_options(self) -> list[str]:
        if not self.options:
            raise RuntimeError
        return [x["value"] for x in self.options]

    def validate(
        self, value: Optional[Union[str, list[str]]]
    ) -> Union[list[str], list[JID], str, JID, bool, None]:
        """
        Raise appropriate XMPPError if a given value is valid for this field

        :param value: The value to test
        :return: The same value OR a JID if ``self.type=jid-single``
        """
        if isinstance(value, list) and not self.type.endswith("multi"):
            raise XMPPError("not-acceptable", "A single value was expected")

        if self.type in ("list-multi", "jid-multi"):
            if not value:
                value = []
            if isinstance(value, list):
                return self.__validate_list_multi(value)
            else:
                raise XMPPError("not-acceptable", "Multiple values was expected")

        assert isinstance(value, (str, bool, JID)) or value is None

        if self.required and value is None:
            raise XMPPError("not-acceptable", f"Missing field: '{self.label}'")

        if value is None:
            return None

        if self.type == "jid-single":
            try:
                return JID(value)
            except ValueError:
                raise XMPPError("not-acceptable", f"Not a valid JID: '{value}'")

        elif self.type == "list-single":
            if value not in self.__acceptable_options():
                raise XMPPError("not-acceptable", f"Not a valid option: '{value}'")

        elif self.type == "boolean":
            return value.lower() in ("1", "true") if isinstance(value, str) else value

        return value

    def __validate_list_multi(self, value: list[str]) -> Union[list[str], list[JID]]:
        for v in value:
            if v not in self.__acceptable_options():
                raise XMPPError("not-acceptable", f"Not a valid option: '{v}'")
        if self.type == "list-multi":
            return value
        return [JID(v) for v in value]

    def get_xml(self) -> SlixFormField:
        """
        Get the field in slixmpp format

        :return: some XML
        """
        f = SlixFormField()
        f["var"] = self.var
        f["label"] = self.label
        f["required"] = self.required
        f["type"] = self.type
        if self.options:
            for o in self.options:
                f.add_option(**o)  # type: ignore[no-untyped-call]
        f["value"] = self.value
        if self.image_url:
            f["media"].add_uri(self.image_url, itype="image/png")
        return f


CommandResponseType = Union[TableResult, Confirmation, Form, str, None]


class Command(ABC):
    """
    Abstract base class to implement gateway commands (chatbot and ad-hoc)
    """

    NAME: str = NotImplemented
    """
    Friendly name of the command, eg: "do something with stuff"
    """
    HELP: str = NotImplemented
    """
    Long description of what the command does
    """
    NODE: str = NotImplemented
    """
    Name of the node used for ad-hoc commands
    """
    CHAT_COMMAND: str = NotImplemented
    """
    Text to send to the gateway to trigger the command via a message
    """

    ACCESS: "CommandAccess" = NotImplemented
    """
    Who can use this command
    """

    CATEGORY: Optional[str] = None
    """
    If used, the command will be under this top-level category.
    Use the same string for several commands to group them.
    This hierarchy only used for the adhoc interface, not the chat command
    interface.
    """

    subclasses = list[Type["Command"]]()

    def __init__(self, xmpp: "BaseGateway"):
        self.xmpp = xmpp

    def __init_subclass__(cls, **kwargs: Any) -> None:
        # store subclasses so subclassing is enough for the command to be
        # picked up by slidge
        cls.subclasses.append(cls)

    async def run(
        self, session: Optional["BaseSession[Any, Any]"], ifrom: JID, *args: str
    ) -> CommandResponseType:
        """
        Entry point of the command

        :param session: If triggered by a registered user, its slidge Session
        :param ifrom: JID of the command-triggering entity
        :param args: When triggered via chatbot type message, additional words
            after the CHAT_COMMAND string was passed

        :return: Either a TableResult, a Form, a Confirmation, a text, or None
        """
        raise XMPPError("feature-not-implemented")

    def _get_session(self, jid: JID) -> Optional["BaseSession[Any, Any]"]:
        user = user_store.get_by_jid(jid)
        if user is None:
            return None

        return self.xmpp.get_session_from_user(user)

    def __can_use_command(self, jid: JID):
        j = jid.bare
        return self.xmpp.jid_validator.match(j) or j in config.ADMINS

    def raise_if_not_authorized(self, jid: JID) -> Optional["BaseSession[Any, Any]"]:
        """
        Raise an appropriate error is jid is not authorized to use the command

        :param jid: jid of the entity trying to access the command
        :return:session of JID if it exists
        """
        session = self._get_session(jid)
        if not self.__can_use_command(jid):
            raise XMPPError(
                "bad-request", "Your JID is not allowed to use this gateway."
            )

        if self.ACCESS == CommandAccess.ADMIN_ONLY and not is_admin(jid):
            raise XMPPError("not-authorized")
        elif self.ACCESS == CommandAccess.NON_USER and session is not None:
            raise XMPPError(
                "bad-request", "This is only available for non-users. Unregister first."
            )
        elif self.ACCESS == CommandAccess.USER and session is None:
            raise XMPPError(
                "forbidden",
                "This is only available for users that are registered to this gateway",
            )
        elif self.ACCESS == CommandAccess.USER_NON_LOGGED:
            if session is None or session.logged:
                raise XMPPError(
                    "forbidden",
                    (
                        "This is only available for users that are not logged to the"
                        " legacy service"
                    ),
                )
        elif self.ACCESS == CommandAccess.USER_LOGGED:
            if session is None or not session.logged:
                raise XMPPError(
                    "forbidden",
                    (
                        "This is only available when you are logged in to the legacy"
                        " service"
                    ),
                )
        return session


def is_admin(jid: JidStr) -> bool:
    return JID(jid).bare in config.ADMINS



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.plugins.base import register_plugin

from .link_preview import LinkPreview

register_plugin(LinkPreview)



import abc
import logging
from typing import TYPE_CHECKING, Generic, Type

from slixmpp import JID
from slixmpp.jid import _unescape_node

from ..contact.roster import ESCAPE_TABLE
from ..core.mixins.lock import NamedLockMixin
from ..util import SubclassableOnce
from ..util.types import LegacyGroupIdType, LegacyMUCType
from .room import LegacyMUC

if TYPE_CHECKING:
    from slidge.core.session import BaseSession


class LegacyBookmarks(
    Generic[LegacyGroupIdType, LegacyMUCType],
    NamedLockMixin,
    metaclass=SubclassableOnce,
):
    """
    This is instantiated once per :class:`~slidge.BaseSession`
    """

    def __init__(self, session: "BaseSession"):
        self.session = session
        self.xmpp = session.xmpp
        self.user = session.user

        self._mucs_by_legacy_id = dict[LegacyGroupIdType, LegacyMUCType]()
        self._mucs_by_bare_jid = dict[str, LegacyMUCType]()

        self._muc_class: Type[LegacyMUCType] = LegacyMUC.get_self_or_unique_subclass()

        self._user_nick: str = self.session.user.jid.node

        super().__init__()
        self.log = logging.getLogger(f"{self.user.bare_jid}:bookmarks")
        self.ready = self.session.xmpp.loop.create_future()
        if not self.xmpp.GROUPS:
            self.ready.set_result(True)

    @property
    def user_nick(self):
        return self._user_nick

    @user_nick.setter
    def user_nick(self, nick: str):
        self._user_nick = nick

    def __iter__(self):
        return iter(self._mucs_by_legacy_id.values())

    def __repr__(self):
        return f"<Bookmarks of {self.user}>"

    async def __finish_init_muc(self, legacy_id: LegacyGroupIdType, jid: JID):
        muc = self._muc_class(self.session, legacy_id=legacy_id, jid=jid)
        await muc.avatar_wrap_update_info()
        if not muc.user_nick:
            muc.user_nick = self._user_nick
        self.log.debug("MUC created: %r", muc)
        self._mucs_by_legacy_id[muc.legacy_id] = muc
        self._mucs_by_bare_jid[muc.jid.bare] = muc
        return muc

    async def legacy_id_to_jid_local_part(self, legacy_id: LegacyGroupIdType):
        return await self.legacy_id_to_jid_username(legacy_id)

    async def jid_local_part_to_legacy_id(self, local_part: str):
        return await self.jid_username_to_legacy_id(local_part)

    async def legacy_id_to_jid_username(self, legacy_id: LegacyGroupIdType):
        """
        The default implementation calls ``str()`` on the legacy_id and
        escape characters according to :xep:`0106`.

        You can override this class and implement a more subtle logic to raise
        an :class:`~slixmpp.exceptions.XMPPError` early

        :param legacy_id:
        :return:
        """
        return str(legacy_id).translate(ESCAPE_TABLE)

    async def jid_username_to_legacy_id(self, username: str):
        """

        :param username:
        :return:
        """
        return _unescape_node(username)

    async def by_jid(self, jid: JID) -> LegacyMUCType:
        bare = jid.bare
        async with self.lock(("bare", bare)):
            muc = self._mucs_by_bare_jid.get(bare)
            if muc is None:
                self.log.debug("Attempting to instantiate a new MUC for JID %s", jid)
                local_part = jid.node
                legacy_id = await self.jid_local_part_to_legacy_id(local_part)
                if self.get_lock(("legacy_id", legacy_id)):
                    self.log.debug("Not instantiating %s after all", jid)
                    return await self.by_legacy_id(legacy_id)
                self.log.debug("%r is group %r", local_part, legacy_id)
                muc = await self.__finish_init_muc(legacy_id, JID(bare))
            else:
                self.log.trace("Found an existing MUC instance: %s", muc)  # type:ignore
            return muc

    async def by_legacy_id(self, legacy_id: LegacyGroupIdType) -> LegacyMUCType:
        async with self.lock(("legacy_id", legacy_id)):
            muc = self._mucs_by_legacy_id.get(legacy_id)
            if muc is None:
                self.log.debug("Create new MUC instance for legacy ID %s", legacy_id)
                local = await self.legacy_id_to_jid_local_part(legacy_id)
                bare = f"{local}@{self.xmpp.boundjid}"
                jid = JID(bare)
                if self.get_lock(("bare", bare)):
                    self.log.debug("Not instantiating %s after all", legacy_id)
                    return await self.by_jid(jid)
                muc = await self.__finish_init_muc(legacy_id, jid)
            else:
                self.log.trace("Found an existing MUC instance: %s", muc)  # type:ignore

            return muc

    @abc.abstractmethod
    async def fill(self):
        """
        Establish a user's known groups.

        This has to be overridden in plugins with group support and at the
        minimum, this should ``await self.by_legacy_id(group_id)`` for all
        the groups a user is part of.

        Slidge internals will call this on successful :meth:`BaseSession.login`

        """
        if self.xmpp.GROUPS:
            raise NotImplementedError(
                "The plugin advertised support for groups but"
                " LegacyBookmarks.fill() was not overridden."
            )

    def remove(self, muc: LegacyMUC):
        try:
            del self._mucs_by_legacy_id[muc.legacy_id]
        except KeyError:
            self.log.warning("Removed a MUC that we didn't store by legacy ID")
        try:
            del self._mucs_by_bare_jid[muc.jid.bare]
        except KeyError:
            self.log.warning("Removed a MUC that we didn't store by JID")
        for part in muc._participants_by_contacts.values():
            try:
                part.contact.participants.remove(part)
            except KeyError:
                part.log.warning(
                    "That participant wasn't stored in the contact's participants attribute"
                )



import asyncio
import functools
import logging
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

from slixmpp import JID, Iq  # type: ignore[attr-defined]
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0004 import Form as SlixForm  # type: ignore[attr-defined]
from slixmpp.plugins.xep_0030.stanza.items import DiscoItems

from . import Command, CommandResponseType, Confirmation, Form, TableResult
from .base import FormField

if TYPE_CHECKING:
    from ..core.gateway.base import BaseGateway
    from ..core.session import BaseSession


AdhocSessionType = dict[str, Any]


class AdhocProvider:
    """
    A slixmpp-like plugin to handle adhoc commands, with less boilerplate and
    untyped dict values than slixmpp.
    """

    def __init__(self, xmpp: "BaseGateway") -> None:
        self.xmpp = xmpp
        self._commands = dict[str, Command]()
        self._categories = dict[str, list[Command]]()
        xmpp.plugin["xep_0030"].set_node_handler(
            "get_items",
            jid=xmpp.boundjid,
            node=self.xmpp.plugin["xep_0050"].stanza.Command.namespace,
            handler=self.get_items,
        )

    async def __wrap_initial_handler(
        self, command: Command, iq: Iq, adhoc_session: AdhocSessionType
    ) -> AdhocSessionType:
        ifrom = iq.get_from()
        session = command.raise_if_not_authorized(ifrom)
        result = await self.__wrap_handler(command.run, session, ifrom)
        return await self.__handle_result(session, result, adhoc_session)

    async def __handle_category_list(
        self, category: str, iq: Iq, adhoc_session: AdhocSessionType
    ) -> AdhocSessionType:
        session = self.xmpp.get_session_from_stanza(iq)
        commands = []
        for command in self._categories[category]:
            try:
                command.raise_if_not_authorized(iq.get_from())
            except XMPPError:
                continue
            commands.append(command)
        return await self.__handle_result(
            session,
            Form(
                category,
                "",
                [
                    FormField(
                        var="command",
                        label="Command",
                        type="list-single",
                        options=[
                            {"label": command.NAME, "value": str(i)}
                            for i, command in enumerate(commands)
                        ],
                    )
                ],
                partial(self.__handle_category_choice, commands),
            ),
            adhoc_session,
        )

    async def __handle_category_choice(
        self,
        commands: list[Command],
        form_values: dict[str, str],
        session: "BaseSession[Any, Any]",
        jid: JID,
    ):
        command = commands[int(form_values["command"])]
        result = await self.__wrap_handler(command.run, session, jid)
        return result

    async def __handle_result(
        self,
        session: Optional["BaseSession[Any, Any]"],
        result: CommandResponseType,
        adhoc_session: AdhocSessionType,
    ) -> AdhocSessionType:
        if isinstance(result, str) or result is None:
            adhoc_session["has_next"] = False
            adhoc_session["next"] = None
            adhoc_session["payload"] = None
            adhoc_session["notes"] = [("info", result or "Success!")]
            return adhoc_session

        if isinstance(result, Form):
            adhoc_session["next"] = partial(self.__wrap_form_handler, session, result)
            adhoc_session["has_next"] = True
            adhoc_session["payload"] = result.get_xml()
            return adhoc_session

        if isinstance(result, Confirmation):
            adhoc_session["next"] = partial(self.__wrap_confirmation, session, result)
            adhoc_session["has_next"] = True
            adhoc_session["payload"] = result.get_form()
            adhoc_session["next"] = partial(self.__wrap_confirmation, session, result)
            return adhoc_session

        if isinstance(result, TableResult):
            adhoc_session["next"] = None
            adhoc_session["has_next"] = False
            adhoc_session["payload"] = result.get_xml()
            return adhoc_session

        raise XMPPError("internal-server-error", text="OOPS!")

    @staticmethod
    async def __wrap_handler(f: Union[Callable, functools.partial], *a, **k):  # type: ignore
        try:
            if asyncio.iscoroutinefunction(f):
                return await f(*a, **k)
            elif hasattr(f, "func") and asyncio.iscoroutinefunction(f.func):
                return await f(*a, **k)
            else:
                return f(*a, **k)
        except Exception as e:
            log.debug("Exception in %s", f, exc_info=e)
            raise XMPPError("internal-server-error", text=str(e))

    async def __wrap_form_handler(
        self,
        session: Optional["BaseSession[Any, Any]"],
        result: Form,
        form: SlixForm,
        adhoc_session: AdhocSessionType,
    ) -> AdhocSessionType:
        form_values = result.get_values(form)
        new_result = await self.__wrap_handler(
            result.handler,
            form_values,
            session,
            adhoc_session["from"],
            *result.handler_args,
            **result.handler_kwargs,
        )

        return await self.__handle_result(session, new_result, adhoc_session)

    async def __wrap_confirmation(
        self,
        session: Optional["BaseSession[Any, Any]"],
        confirmation: Confirmation,
        form: SlixForm,
        adhoc_session: AdhocSessionType,
    ) -> AdhocSessionType:
        if form.get_values().get("confirm"):  # type: ignore[no-untyped-call]
            result = await self.__wrap_handler(
                confirmation.handler,
                session,
                adhoc_session["from"],
                *confirmation.handler_args,
                **confirmation.handler_kwargs,
            )
            if confirmation.success:
                result = confirmation.success
        else:
            result = "You canceled the operation"

        return await self.__handle_result(session, result, adhoc_session)

    def register(self, command: Command, jid: Optional[JID] = None) -> None:
        """
        Register a command as a adhoc command.

        this does not need to be called manually, ``BaseGateway`` takes care of
        that.

        :param command:
        :param jid:
        """
        if jid is None:
            jid = self.xmpp.boundjid
        elif not isinstance(jid, JID):
            jid = JID(jid)

        if (category := command.CATEGORY) is None:
            if command.NODE in self._commands:
                raise RuntimeError(
                    "There is already a command for the node '%s'", command.NODE
                )
            self._commands[command.NODE] = command
            self.xmpp.plugin["xep_0050"].add_command(  # type: ignore[no-untyped-call]
                jid=jid,
                node=command.NODE,
                name=command.NAME,
                handler=partial(self.__wrap_initial_handler, command),
            )
        else:
            if category not in self._categories:
                self._categories[category] = list[Command]()
                self.xmpp.plugin["xep_0050"].add_command(  # type: ignore[no-untyped-call]
                    jid=jid,
                    node=category,
                    name=category,
                    handler=partial(self.__handle_category_list, category),
                )
            self._categories[category].append(command)

    async def get_items(self, jid: JID, node: str, iq: Iq) -> DiscoItems:
        """
        Get items for a disco query

        :param jid: who is requesting the disco
        :param node: which command node is requested
        :param iq:  the disco query IQ
        :return: commands accessible to the given JID will be listed
        """
        all_items = self.xmpp.plugin["xep_0030"].static.get_items(jid, node, None, None)
        log.debug("Static items: %r", all_items)
        if not all_items:
            return DiscoItems()

        ifrom = iq.get_from()

        filtered_items = DiscoItems()
        filtered_items["node"] = self.xmpp.plugin["xep_0050"].stanza.Command.namespace
        for item in all_items:
            authorized = True
            if item["node"] in self._categories:
                for command in self._categories[item["node"]]:
                    try:
                        command.raise_if_not_authorized(ifrom)
                    except XMPPError:
                        authorized = False
                    else:
                        authorized = True
                        break
            else:
                try:
                    self._commands[item["node"]].raise_if_not_authorized(ifrom)
                except XMPPError:
                    authorized = False

            if authorized:
                filtered_items.append(item)

        return filtered_items


log = logging.getLogger(__name__)



import asyncio
import hashlib
import io
import logging
import shelve
import uuid
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from http import HTTPStatus
from pathlib import Path
from typing import Optional

import aiohttp
from multidict import CIMultiDictProxy
from PIL import Image
from slixmpp import JID

from ..util.types import URL, LegacyFileIdType
from . import config


@dataclass
class CachedAvatar:
    filename: str
    hash: str
    height: int
    width: int
    root: Path
    etag: Optional[str] = None
    last_modified: Optional[str] = None

    @property
    def data(self):
        return self.path.read_bytes()

    @property
    def path(self):
        return self.root / self.filename


class AvatarCache:
    _shelf_path: str
    _jid_to_legacy_path: str
    dir: Path
    http: aiohttp.ClientSession

    def __init__(self):
        self._thread_pool = ThreadPoolExecutor(config.AVATAR_RESAMPLING_THREADS)

    def set_dir(self, path: Path):
        self.dir = path
        self.dir.mkdir(exist_ok=True)
        self._shelf_path = str(path / "slidge_avatar_cache.shelf")
        self._jid_to_legacy_path = str(path / "jid_to_avatar_unique_id.shelf")

    def close(self):
        self._thread_pool.shutdown(cancel_futures=True)

    def __get_http_headers(self, cached: Optional[CachedAvatar]):
        headers = {}
        if cached and (self.dir / cached.filename).exists():
            if last_modified := cached.last_modified:
                headers["If-Modified-Since"] = last_modified
            if etag := cached.etag:
                headers["If-None-Match"] = etag
        return headers

    async def get_avatar_from_url_alone(self, url: str, jid: JID):
        """
        Used when no avatar unique ID is passed. Store and use http headers
        to avoid fetching ut
        """
        cached = self.get(url)
        headers = self.__get_http_headers(cached)
        async with _download_lock:
            return await self.__download(cached, url, headers, jid)

    async def __download(
        self,
        cached: Optional[CachedAvatar],
        url: str,
        headers: dict[str, str],
        jid: JID,
    ):
        async with self.http.get(url, headers=headers) as response:
            if response.status == HTTPStatus.NOT_MODIFIED:
                log.debug("Using avatar cache for %s", jid)
                return cached
            log.debug("Download avatar for %s", jid)
            return await self.convert_and_store(
                Image.open(io.BytesIO(await response.read())),
                url,
                jid,
                response.headers,
            )

    async def url_has_changed(self, url: URL):
        with shelve.open(self._shelf_path) as s:
            cached = s.get(url)
        if cached is None:
            return True
        headers = self.__get_http_headers(cached)
        async with self.http.head(url, headers=headers) as response:
            return response.status != HTTPStatus.NOT_MODIFIED

    def get(self, unique_id: LegacyFileIdType) -> Optional[CachedAvatar]:
        with shelve.open(self._shelf_path) as s:
            return s.get(str(unique_id))

    def get_cached_id_for(self, jid: JID) -> Optional[LegacyFileIdType]:
        with shelve.open(self._jid_to_legacy_path) as s:
            return s.get(str(jid))

    def store_jid(self, jid: JID, uid: LegacyFileIdType):
        with shelve.open(self._jid_to_legacy_path) as s:
            s[str(jid)] = uid

    def delete_jid(self, jid: JID):
        try:
            with shelve.open(self._jid_to_legacy_path) as s:
                del s[str(jid)]
        except KeyError:
            pass

    async def convert_and_store(
        self,
        img: Image.Image,
        unique_id: LegacyFileIdType,
        jid: JID,
        response_headers: Optional[CIMultiDictProxy[str]] = None,
    ) -> CachedAvatar:
        resize = (size := config.AVATAR_SIZE) and any(x > size for x in img.size)
        if resize:
            await asyncio.get_event_loop().run_in_executor(
                self._thread_pool, img.thumbnail, (size, size)
            )
            log.debug("Resampled image to %s", img.size)

        filename = str(uuid.uuid1()) + ".png"
        file_path = self.dir / filename

        if (
            not resize
            and img.format == "PNG"
            and isinstance(unique_id, str)
            and (path := Path(unique_id))
            and path.exists()
        ):
            img_bytes = path.read_bytes()
        else:
            with io.BytesIO() as f:
                img.save(f, format="PNG")
                img_bytes = f.getvalue()

        with file_path.open("wb") as file:
            file.write(img_bytes)

        hash_ = hashlib.sha1(img_bytes).hexdigest()

        avatar = CachedAvatar(
            filename=filename,
            hash=hash_,
            height=img.height,
            width=img.width,
            root=self.dir,
        )
        if response_headers:
            avatar.etag = response_headers.get("etag")
            avatar.last_modified = response_headers.get("last-modified")
        with shelve.open(self._shelf_path) as s:
            s[str(unique_id)] = avatar
        self.store_jid(jid, unique_id)
        return avatar


avatar_cache = AvatarCache()
log = logging.getLogger(__name__)
_download_lock = asyncio.Lock()

__all__ = (
    "CachedAvatar",
    "avatar_cache",
)



from asyncio import Task, create_task
from hashlib import sha1
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from slixmpp import JID

from ...util.types import (
    URL,
    AnyBaseSession,
    AvatarIdType,
    AvatarType,
    LegacyFileIdType,
)
from ..cache import avatar_cache

if TYPE_CHECKING:
    from ..pubsub import PepAvatar


class AvatarMixin:
    """
    Mixin for XMPP entities that have avatars that represent them.

    Both :py:class:`slidge.LegacyContact` and :py:class:`slidge.LegacyMUC` use
    :py:class:`.AvatarMixin`.
    """

    jid: JID = NotImplemented
    session: AnyBaseSession = NotImplemented
    _avatar_pubsub_broadcast: bool = NotImplemented
    _avatar_bare_jid: bool = NotImplemented

    def __init__(self) -> None:
        super().__init__()
        self._set_avatar_task: Optional[Task] = None
        self.__avatar_unique_id: Optional[AvatarIdType] = None

    @property
    def __avatar_jid(self):
        return JID(self.jid.bare) if self._avatar_bare_jid else self.jid

    @property
    def avatar_id(self) -> Optional[AvatarIdType]:
        """
        The unique ID of this entity's avatar.
        """
        return self.__avatar_unique_id

    @property
    def avatar(self) -> Optional[AvatarIdType]:
        """
        This property can be used to set the avatar, but
        :py:meth:`~.AvatarMixin.set_avatar()` should be preferred because you can
        provide a unique ID for the avatar for efficient caching.
        Setting this is OKish in case the avatar type is a URL or a local path
        that can act as a legacy ID.

        Python's ``property`` is abused here to maintain backwards
        compatibility, but when getting it you actually get the avatar legacy
        ID.
        """
        return self.__avatar_unique_id

    @avatar.setter
    def avatar(self, a: Optional[AvatarType]):
        if self._set_avatar_task:
            self._set_avatar_task.cancel()
        self.session.log.debug("Setting avatar with property")
        self._set_avatar_task = self.session.xmpp.loop.create_task(
            self.set_avatar(a, None, blocking=True, cancel=False),
            name=f"Set avatar of {self} from property",
        )

    @staticmethod
    def __get_uid(a: Optional[AvatarType]) -> Optional[AvatarIdType]:
        if isinstance(a, str):
            return URL(a)
        elif isinstance(a, Path):
            return str(a)
        elif isinstance(a, bytes):
            return sha1(a).hexdigest()
        elif a is None:
            return None
        raise TypeError("Bad avatar", a)

    async def __set_avatar(self, a: Optional[AvatarType], uid: Optional[AvatarIdType]):
        self.__avatar_unique_id = uid
        await self.session.xmpp.pubsub.set_avatar(
            jid=self.__avatar_jid,
            avatar=a,
            unique_id=None if isinstance(uid, URL) else uid,
            broadcast_to=self.session.user.jid.bare,
            broadcast=self._avatar_pubsub_broadcast,
        )
        self._post_avatar_update()

    async def _no_change(self, a: Optional[AvatarType], uid: Optional[AvatarIdType]):
        if a is None:
            return self.__avatar_unique_id is None
        if not self.__avatar_unique_id:
            return False
        if isinstance(uid, URL):
            if self.__avatar_unique_id != uid:
                return False
            return not await avatar_cache.url_has_changed(uid)
        return self.__avatar_unique_id == uid

    async def set_avatar(
        self,
        a: Optional[AvatarType],
        avatar_unique_id: Optional[LegacyFileIdType] = None,
        blocking=False,
        cancel=True,
    ) -> None:
        """
        Set an avatar for this entity

        :param a:
        :param avatar_unique_id:
        :param blocking:
        :param cancel:
        """
        if avatar_unique_id is None and a is not None:
            avatar_unique_id = self.__get_uid(a)
        if await self._no_change(a, avatar_unique_id):
            return
        if cancel and self._set_avatar_task:
            self._set_avatar_task.cancel()
        awaitable = create_task(
            self.__set_avatar(a, avatar_unique_id),
            name=f"Set pubsub avatar of {self}",
        )
        if not self._set_avatar_task or self._set_avatar_task.done():
            self._set_avatar_task = awaitable
        if blocking:
            await awaitable

    def get_avatar(self) -> Optional["PepAvatar"]:
        if not self.__avatar_unique_id:
            return None
        return self.session.xmpp.pubsub.get_avatar(self.__avatar_jid)

    def _post_avatar_update(self) -> None:
        return

    async def avatar_wrap_update_info(self):
        cached_id = avatar_cache.get_cached_id_for(self.__avatar_jid)
        self.__avatar_unique_id = cached_id
        try:
            await self.update_info()  # type:ignore
        except NotImplementedError:
            return
        new_id = self.avatar
        if isinstance(new_id, URL) and not await avatar_cache.url_has_changed(new_id):
            return
        elif new_id != cached_id:
            # at this point it means that update_info set the avatar, and we don't
            # need to do anything else
            return

        await self.session.xmpp.pubsub.set_avatar_from_cache(
            self.__avatar_jid,
            new_id is None and cached_id is not None,
            self.session.user.jid.bare,
            self._avatar_pubsub_broadcast,
        )



import re
from asyncio import Task, sleep
from datetime import datetime, timedelta, timezone
from typing import Optional

from slixmpp.types import PresenceShows, PresenceTypes

from ...util.sql import CachedPresence, db
from .. import config
from .base import BaseSender


class _NoChange(Exception):
    pass


_FRIEND_REQUEST_PRESENCES = {"subscribe", "unsubscribe", "subscribed", "unsubscribed"}


class PresenceMixin(BaseSender):
    _ONLY_SEND_PRESENCE_CHANGES = False

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.__update_last_seen_fallback_task: Optional[Task] = None

    async def __update_last_seen_fallback(self):
        await sleep(3600 * 7)
        self.send_last_presence(force=True, no_cache_online=False)

    def _get_last_presence(self) -> Optional[CachedPresence]:
        return db.presence_get(self.jid, self.user)

    def _store_last_presence(self, new: CachedPresence):
        return db.presence_store(self.jid, new, self.user)

    def _make_presence(
        self,
        *,
        last_seen: Optional[datetime] = None,
        force=False,
        bare=False,
        ptype: Optional[PresenceTypes] = None,
        pstatus: Optional[str] = None,
        pshow: Optional[PresenceShows] = None,
    ):
        if last_seen and last_seen.tzinfo is None:
            last_seen = last_seen.astimezone(timezone.utc)

        old = self._get_last_presence()

        if ptype not in _FRIEND_REQUEST_PRESENCES:
            new = CachedPresence(
                last_seen=last_seen, ptype=ptype, pstatus=pstatus, pshow=pshow
            )
            if old != new:
                if hasattr(self, "muc") and ptype == "unavailable":
                    db.presence_delete(self.jid, self.user)
                else:
                    self._store_last_presence(new)
            if old and not force and self._ONLY_SEND_PRESENCE_CHANGES:
                if old == new:
                    self.session.log.debug("Presence is the same as cached")
                    raise _NoChange
                self.session.log.debug(
                    "Presence is not the same as cached: %s vs %s", old, new
                )

        p = self.xmpp.make_presence(
            pfrom=self.jid.bare if bare else self.jid,
            ptype=ptype,
            pshow=pshow,
            pstatus=pstatus,
        )
        if last_seen:
            # it's ugly to check for the presence of this string, but a better fix is more work
            if config.LAST_SEEN_FALLBACK and not re.match(
                ".*Last seen .*", p["status"]
            ):
                last_seen_fallback, recent = get_last_seen_fallback(last_seen)
                if p["status"]:
                    p["status"] = p["status"] + " -- " + last_seen_fallback
                else:
                    p["status"] = last_seen_fallback
                if recent:
                    # if less than a week, we use sth like 'Last seen: Monday, 8:05",
                    # but if lasts more than a week, this is not very informative, so
                    # we need to force resend an updated presence status
                    if self.__update_last_seen_fallback_task:
                        self.__update_last_seen_fallback_task.cancel()
                    self.__update_last_seen_fallback_task = self.xmpp.loop.create_task(
                        self.__update_last_seen_fallback()
                    )
            p["idle"]["since"] = last_seen
        return p

    def send_last_presence(self, force=False, no_cache_online=False):
        if (cache := self._get_last_presence()) is None:
            if force:
                if no_cache_online:
                    self.online()
                else:
                    self.offline()
            return
        self._send(
            self._make_presence(
                last_seen=cache.last_seen,
                force=True,
                ptype=cache.ptype,
                pshow=cache.pshow,
                pstatus=cache.pstatus,
            )
        )

    def online(
        self,
        status: Optional[str] = None,
        last_seen: Optional[datetime] = None,
    ):
        """
        Send an "online" presence from this contact to the user.

        :param status: Arbitrary text, details of the status, eg: "Listening to Britney Spears"
        :param last_seen: For :xep:`0319`
        """
        try:
            self._send(self._make_presence(pstatus=status, last_seen=last_seen))
        except _NoChange:
            pass

    def away(
        self,
        status: Optional[str] = None,
        last_seen: Optional[datetime] = None,
    ):
        """
        Send an "away" presence from this contact to the user.

        This is a global status, as opposed to :meth:`.LegacyContact.inactive`
        which concerns a specific conversation, ie a specific "chat window"

        :param status: Arbitrary text, details of the status, eg: "Gone to fight capitalism"
        :param last_seen: For :xep:`0319`
        """
        try:
            self._send(
                self._make_presence(pstatus=status, pshow="away", last_seen=last_seen)
            )
        except _NoChange:
            pass

    def extended_away(
        self,
        status: Optional[str] = None,
        last_seen: Optional[datetime] = None,
    ):
        """
        Send an "extended away" presence from this contact to the user.

        This is a global status, as opposed to :meth:`.LegacyContact.inactive`
        which concerns a specific conversation, ie a specific "chat window"

        :param status: Arbitrary text, details of the status, eg: "Gone to fight capitalism"
        :param last_seen: For :xep:`0319`
        """
        try:
            self._send(
                self._make_presence(pstatus=status, pshow="xa", last_seen=last_seen)
            )
        except _NoChange:
            pass

    def busy(
        self,
        status: Optional[str] = None,
        last_seen: Optional[datetime] = None,
    ):
        """
        Send a "busy" (ie, "dnd") presence from this contact to the user,

        :param status: eg: "Trying to make sense of XEP-0100"
        :param last_seen: For :xep:`0319`
        """
        try:
            self._send(
                self._make_presence(pstatus=status, pshow="dnd", last_seen=last_seen)
            )
        except _NoChange:
            pass

    def offline(
        self,
        status: Optional[str] = None,
        last_seen: Optional[datetime] = None,
    ):
        """
        Send an "offline" presence from this contact to the user.

        :param status: eg: "Trying to make sense of XEP-0100"
        :param last_seen: For :xep:`0319`
        """
        try:
            self._send(
                self._make_presence(
                    pstatus=status, ptype="unavailable", last_seen=last_seen
                )
            )
        except _NoChange:
            pass


def get_last_seen_fallback(last_seen: datetime):
    now = datetime.now(tz=timezone.utc)
    if now - last_seen < timedelta(days=7):
        return f"Last seen {last_seen:%A %H:%M GMT}", True
    else:
        return f"Last seen {last_seen:%b %-d %Y}", False



import logging
import uuid
from copy import copy
from datetime import datetime
from typing import TYPE_CHECKING, Collection, Optional

from slixmpp import Iq, Message

from ..util.archive_msg import HistoryMessage
from ..util.db import GatewayUser
from ..util.sql import db

if TYPE_CHECKING:
    from .participant import LegacyParticipant


class MessageArchive:
    def __init__(self, db_id: str, user: GatewayUser):
        self.db_id = db_id
        self.user = user
        db.mam_add_muc(db_id, user)

    def add(
        self,
        msg: Message,
        participant: Optional["LegacyParticipant"] = None,
    ):
        """
        Add a message to the archive if it is deemed archivable

        :param msg:
        :param participant:
        """
        if not archivable(msg):
            return
        new_msg = copy(msg)
        if participant and not participant.muc.is_anonymous:
            new_msg["muc"]["role"] = participant.role
            new_msg["muc"]["affiliation"] = participant.affiliation
            if participant.contact:
                new_msg["muc"]["jid"] = participant.contact.jid.bare
            elif participant.is_user:
                new_msg["muc"]["jid"] = participant.user.jid.bare
            elif participant.is_system:
                new_msg["muc"]["jid"] = participant.muc.jid
            else:
                log.warning("No real JID for participant in this group")
                new_msg["muc"][
                    "jid"
                ] = f"{uuid.uuid4()}@{participant.xmpp.boundjid.bare}"

        db.mam_add_msg(self.db_id, HistoryMessage(new_msg), self.user)

    def __iter__(self):
        return iter(self.get_all())

    def get_all(
        self,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        before_id: Optional[str] = None,
        after_id: Optional[str] = None,
        ids: Collection[str] = (),
        last_page_n: Optional[int] = None,
        sender: Optional[str] = None,
        flip=False,
    ):
        for msg in db.mam_get_messages(
            self.user,
            self.db_id,
            before_id=before_id,
            after_id=after_id,
            ids=ids,
            last_page_n=last_page_n,
            sender=sender,
            start_date=start_date,
            end_date=end_date,
            flip=flip,
        ):
            yield msg

    async def send_metadata(self, iq: Iq):
        """
        Send archive extent, as per the spec

        :param iq:
        :return:
        """
        reply = iq.reply()
        messages = db.mam_get_first_and_last(self.db_id)
        if messages:
            for x, m in [("start", messages[0]), ("end", messages[-1])]:
                reply["mam_metadata"][x]["id"] = m.id
                reply["mam_metadata"][x]["timestamp"] = m.sent_on
        else:
            reply.enable("mam_metadata")
        reply.send()


def archivable(msg: Message):
    """
    Determine if a message stanza is worth archiving, ie, convey meaningful
    info

    :param msg:
    :return:
    """

    if msg.get_plugin("hint", check=True) and msg["hint"] == "no-store":
        return False

    if msg["body"]:
        return True

    if msg.get_plugin("apply_to", check=True):
        # retractions
        return True

    if msg.get_plugin("reactions", check=True):
        return True

    return False


log = logging.getLogger(__name__)



import logging
import warnings

from slixmpp import JID, Iq, Message, Presence, register_stanza_plugin
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.base import BasePlugin

from slidge.core import config

from . import stanza

log = logging.getLogger(__name__)


class XEP_0100(BasePlugin):
    name = "xep_0100"
    description = "XEP-0100: Gateway interaction (slidge)"
    dependencies = {
        "xep_0030",  # Service discovery
        "xep_0077",  # In band registration
        "xep_0356",  # Privileged entities
    }

    default_config = {
        "component_name": "SliXMPP gateway",
        "type": "xmpp",
        "needs_registration": True,
    }

    def plugin_init(self):
        if not self.xmpp.is_component:
            log.error("Only components can be gateways, aborting plugin load")
            return

        self.xmpp["xep_0030"].add_identity(
            name=self.component_name, category="gateway", itype=self.type
        )

        # Without that BaseXMPP sends unsub/unavailable on sub requests, and we don't want that
        self.xmpp.client_roster.auto_authorize = False
        self.xmpp.client_roster.auto_subscribe = False

        self.xmpp.add_event_handler("user_register", self.on_user_register)
        self.xmpp.add_event_handler("user_unregister", self.on_user_unregister)
        self.xmpp.add_event_handler(
            "presence_unsubscribe", self.on_presence_unsubscribe
        )

        self.xmpp.add_event_handler("message", self.on_message)

        register_stanza_plugin(Iq, stanza.Gateway)

    def plugin_end(self):
        if not self.xmpp.is_component:
            self.xmpp.remove_event_handler("user_register", self.on_user_register)
            self.xmpp.remove_event_handler("user_unregister", self.on_user_unregister)
            self.xmpp.remove_event_handler(
                "presence_unsubscribe", self.on_presence_unsubscribe
            )

            self.xmpp.remove_event_handler("message", self.on_message)

    async def get_user(self, stanza):
        return await self.xmpp["xep_0077"].api["user_get"](None, None, None, stanza)

    async def on_user_unregister(self, iq: Iq):
        self.xmpp.send_presence(pto=iq.get_from().bare, ptype="unavailable")
        self.xmpp.send_presence(pto=iq.get_from().bare, ptype="unsubscribe")
        self.xmpp.send_presence(pto=iq.get_from().bare, ptype="unsubscribed")

    async def on_user_register(self, iq: Iq):
        self.xmpp.client_roster[iq.get_from()].load()
        await self.add_component_to_roster(jid=iq.get_from())

    async def add_component_to_roster(self, jid: JID):
        if config.NO_ROSTER_PUSH:
            return
        items = {
            self.xmpp.boundjid.bare: {
                "name": self.component_name,
                "subscription": "both",
                "groups": ["Slidge"],
            }
        }
        try:
            await self._set_roster(jid, items)
        except PermissionError:
            warnings.warn(
                "Slidge does not have the privilege to manage users' rosters. "
                "Users should add the slidge component to their rosters manually."
            )
            if config.ROSTER_PUSH_PRESENCE_SUBSCRIPTION_REQUEST_FALLBACK:
                self.xmpp.send_presence(ptype="subscribe", pto=jid.bare)

    async def _set_roster(self, jid, items):
        try:
            await self.xmpp["xep_0356"].set_roster(jid=jid.bare, roster_items=items)
        except PermissionError:
            await self.xmpp["xep_0356_old"].set_roster(jid=jid.bare, roster_items=items)

    def on_presence_unsubscribe(self, p: Presence):
        if p.get_to() == self.xmpp.boundjid.bare:
            log.debug("REMOVE: Our roster: %s", self.xmpp.client_roster)
            self.xmpp["xep_0077"].api["user_remove"](None, None, p["from"], p)
            self.xmpp.event("user_unregister", p)

    async def on_message(self, msg: Message):
        if msg["type"] == "groupchat":
            return  # groupchat messages are out of scope of XEP-0100

        if msg["to"] == self.xmpp.boundjid.bare:
            # It may be useful to exchange direct messages with the component
            self.xmpp.event("gateway_message", msg)
            return

        if self.needs_registration and await self.get_user(msg) is None:
            raise XMPPError(
                "registration-required", text="You are not registered to this gateway"
            )

        self.xmpp.event("legacy_message", msg)
